from __future__ import annotations

import datetime
from collections.abc import Mapping, Iterable, Set
from typing import (
    TYPE_CHECKING,
    Optional,
    assert_never,
    assert_type,
    Literal,
)

import attrs
import cattrs
import numpy as np
import sqlalchemy.orm
from sqlalchemy import func, select
from sqlalchemy.orm import Session

from caqtus.types.data import Data
from caqtus.types.data import is_data, DataLabel
from caqtus.types.iteration import (
    IterationConfiguration,
    Unknown,
)
from caqtus.types.parameter import Parameter, ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.units import Quantity
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils import serialization
from caqtus.utils.result import (
    Success,
    Failure,
    is_failure_type,
    is_success,
    is_failure,
    unwrap,
)
from ._path_hierarchy import _query_path_model
from ._path_table import SQLSequencePath
from ._sequence_table import (
    SQLSequence,
    SQLIterationConfiguration,
    SQLTimelanes,
    SQLDeviceConfiguration,
    SQLSequenceParameters,
    SQLExceptionTraceback,
)
from ._serializer import SerializerProtocol
from ._shot_tables import SQLShot, SQLShotParameter, SQLShotArray, SQLStructuredShotData
from .._data_id import DataId
from .._exception_summary import TracebackSummary
from .._path import PureSequencePath
from .._path_hierarchy import PathNotFoundError, PathHasChildrenError, PathIsRootError
from .._sequence_collection import (
    PathIsSequenceError,
    PathIsNotSequenceError,
    SequenceNotEditableError,
    SequenceStats,
    ShotNotFoundError,
    DataNotFoundError,
    SequenceNotCrashedError,
    InvalidStateTransitionError,
    SequenceNotRunningError,
    SequenceNotLaunchedError,
)
from .._sequence_collection import SequenceCollection
from .._shot_id import ShotId
from .._state import State
from ...device import DeviceName, DeviceConfiguration

if TYPE_CHECKING:
    from ._experiment_session import SQLExperimentSession


@attrs.frozen
class SQLSequenceCollection(SequenceCollection):
    _parent_session: "SQLExperimentSession"
    serializer: SerializerProtocol

    @property
    def parent_session(self) -> "SQLExperimentSession":
        return self._parent_session

    def is_sequence(
        self, path: PureSequencePath
    ) -> Success[bool] | Failure[PathNotFoundError]:
        return _is_sequence(self._get_sql_session(), path)

    def get_contained_sequences(
        self, path: PureSequencePath
    ) -> Success[set[PureSequencePath]] | Failure[PathNotFoundError]:
        path_result = _query_path_model(self._get_sql_session(), path)
        if is_failure_type(path_result, PathNotFoundError):
            return path_result

        result = set()
        if is_failure_type(path_result, PathIsRootError):
            ancestor_id = None
        else:
            ancestor_id = path_result.value.id_

            if path_result.value.sequence is not None:
                result.add(path)

        sequences_query = self.descendant_sequences(ancestor_id)

        query = self._get_sql_session().execute(sequences_query).scalars().all()
        return Success({PureSequencePath(row.path.path) for row in query} | result)

    def descendant_sequences(
        self, ancestor_id: Optional[int]
    ) -> sqlalchemy.sql.Select[tuple[SQLSequence]]:
        """Returns a query for the descendant sequences of the given ancestor.

        The ancestor is not included in the result.
        """

        descendants = self.parent_session.paths.descendants_query(ancestor_id)
        sequences_query = select(SQLSequence).join(
            descendants,
            SQLSequence.path_id == descendants.id_,
        )
        return sequences_query

    def get_contained_running_sequences(
        self, path: PureSequencePath
    ) -> Success[set[PureSequencePath]] | Failure[PathNotFoundError]:
        path_model_result = _query_path_model(self._get_sql_session(), path)

        running_sequences = set()
        if isinstance(path_model_result, Failure):
            if is_failure_type(path_model_result, PathNotFoundError):
                return path_model_result
            assert_type(path_model_result, Failure[PathIsRootError])
            parent_id = None
        else:
            path_model = path_model_result.value

            if path_model.sequence is not None:
                if path_model.sequence.state in {State.PREPARING, State.RUNNING}:
                    running_sequences.add(path)
            parent_id = path_model.id_

        sequences_query = self.descendant_sequences(parent_id)
        running_sequences_query = sequences_query.where(
            SQLSequence.state.in_({State.PREPARING, State.RUNNING})
        )

        result = (
            self._get_sql_session().execute(running_sequences_query).scalars().all()
        )
        running_sequences.update(PureSequencePath(row.path.path) for row in result)
        return Success(running_sequences)

    def _set_global_parameters(
        self, path: PureSequencePath, parameters: ParameterNamespace
    ) -> None:
        sequence = unwrap(self._query_sequence_model(path))
        if sequence.state != State.PREPARING:
            raise SequenceNotEditableError(path)

        if not isinstance(parameters, ParameterNamespace):
            raise TypeError(
                f"Invalid parameters type {type(parameters)}, "
                f"expected ParameterNamespace"
            )

        parameters_content = serialization.converters["json"].unstructure(
            parameters, ParameterNamespace
        )

        sequence.parameters.content = parameters_content

    def get_global_parameters(
        self, path: PureSequencePath
    ) -> (
        Success[ParameterNamespace]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotLaunchedError]
    ):
        return _get_sequence_global_parameters(self._get_sql_session(), path)

    def get_iteration_configuration(
        self, sequence: PureSequencePath
    ) -> IterationConfiguration:
        return _get_iteration_configuration(
            self._get_sql_session(), sequence, self.serializer
        )

    def set_iteration_configuration(
        self,
        sequence: PureSequencePath,
        iteration_configuration: IterationConfiguration,
    ) -> (
        None
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotEditableError]
    ):
        return _set_iteration_configuration(
            self._get_sql_session(), sequence, iteration_configuration, self.serializer
        )

    def create(
        self,
        path: PureSequencePath,
        iteration_configuration: IterationConfiguration,
        time_lanes: TimeLanes,
    ) -> Success[None] | Failure[PathIsSequenceError] | Failure[PathHasChildrenError]:
        children_result = self.parent_session.paths.get_children(path)
        if is_success(children_result):
            if children_result.value:
                return Failure(PathHasChildrenError(path))
        else:
            if is_failure_type(children_result, PathIsSequenceError):
                return children_result
            elif is_failure_type(children_result, PathNotFoundError):
                creation_result = self.parent_session.paths.create_path(path)
                if is_failure(creation_result):
                    if is_failure_type(creation_result, PathIsSequenceError):
                        return creation_result
                    assert_never(creation_result)
            else:
                assert_never(children_result)

        iteration_content = self.serializer.dump_sequence_iteration(
            iteration_configuration
        )

        new_sequence = SQLSequence(
            path=unwrap(self._query_path_model(path)),
            parameters=SQLSequenceParameters(content=None),
            iteration=SQLIterationConfiguration(content=iteration_content),
            time_lanes=SQLTimelanes(content=self.serialize_time_lanes(time_lanes)),
            state=State.DRAFT,
            device_configurations=[],
            start_time=None,
            stop_time=None,
            expected_number_of_shots=_convert_from_unknown(
                iteration_configuration.expected_number_shots()
            ),
        )
        self._get_sql_session().add(new_sequence)
        return Success(None)

    def serialize_time_lanes(self, time_lanes: TimeLanes) -> serialization.JSON:
        return self.serializer.unstructure_time_lanes(time_lanes)

    def get_time_lanes(
        self, sequence_path: PureSequencePath
    ) -> TimeLanes | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]:
        return _get_time_lanes(self._get_sql_session(), sequence_path, self.serializer)

    def set_time_lanes(
        self, sequence_path: PureSequencePath, time_lanes: TimeLanes
    ) -> (
        None
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotEditableError]
    ):
        return _set_time_lanes(
            self._get_sql_session(), sequence_path, time_lanes, self.serializer
        )

    def get_state(
        self, path: PureSequencePath
    ) -> Success[State] | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]:
        result = self._query_sequence_model(path)
        return result.map(lambda sequence: sequence.state)

    def get_exception(
        self, path: PureSequencePath
    ) -> (
        Success[Optional[TracebackSummary]]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotCrashedError]
    ):
        return _get_exceptions(self._get_sql_session(), path)

    def _set_exception(
        self, path: PureSequencePath, exception: TracebackSummary
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotCrashedError]
    ):
        return _set_exception(self._get_sql_session(), path, exception)

    def set_preparing(
        self,
        path: PureSequencePath,
        device_configurations: Mapping[DeviceName, DeviceConfiguration],
        global_parameters: ParameterNamespace,
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        sequence_result = _query_sequence_model(self._get_sql_session(), path)
        if is_failure(sequence_result):
            return sequence_result
        sequence = sequence_result.value
        if not State.is_transition_allowed(sequence.state, State.PREPARING):
            return Failure(
                InvalidStateTransitionError(
                    f"Sequence at {path} can't transition from {sequence.state} to "
                    f"{State.PREPARING}"
                )
            )
        sequence.state = State.PREPARING
        self._set_device_configurations(path, device_configurations)
        self._set_global_parameters(path, global_parameters)
        return Success(None)

    def set_running(
        self, path: PureSequencePath, start_time: datetime.datetime | Literal["now"]
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        sequence_result = _query_sequence_model(self._get_sql_session(), path)
        if is_failure(sequence_result):
            return sequence_result
        sequence = sequence_result.content()
        if not State.is_transition_allowed(sequence.state, State.RUNNING):
            return Failure(
                InvalidStateTransitionError(
                    f"Sequence at {path} can't transition from {sequence.state} to "
                    f"{State.RUNNING}"
                )
            )
        sequence.state = State.RUNNING
        if start_time == "now":
            start_time = datetime.datetime.now(tz=datetime.timezone.utc)
        if not is_tz_aware(start_time):
            raise ValueError("Start time must be timezone aware")
        sequence.start_time = start_time.astimezone(datetime.timezone.utc).replace(
            tzinfo=None
        )
        return Success(None)

    def set_finished(
        self, path: PureSequencePath, stop_time: datetime.datetime | Literal["now"]
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        sequence_result = _query_sequence_model(self._get_sql_session(), path)
        if is_failure(sequence_result):
            return sequence_result
        sequence = sequence_result.value
        if not State.is_transition_allowed(sequence.state, State.FINISHED):
            return Failure(
                InvalidStateTransitionError(
                    f"Sequence at {path} can't transition from {sequence.state} to "
                    f"{State.FINISHED}"
                )
            )
        sequence.state = State.FINISHED
        if stop_time == "now":
            stop_time = datetime.datetime.now(tz=datetime.timezone.utc)
        if not is_tz_aware(stop_time):
            raise ValueError("Stop time must be timezone aware")
        sequence.stop_time = stop_time.astimezone(datetime.timezone.utc).replace(
            tzinfo=None
        )
        return Success(None)

    def set_interrupted(
        self, path: PureSequencePath, stop_time: datetime.datetime | Literal["now"]
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        sequence_result = _query_sequence_model(self._get_sql_session(), path)
        if is_failure(sequence_result):
            return sequence_result
        sequence = sequence_result.value
        if not State.is_transition_allowed(sequence.state, State.INTERRUPTED):
            return Failure(
                InvalidStateTransitionError(
                    f"Sequence at {path} can't transition from {sequence.state} to "
                    f"{State.INTERRUPTED}"
                )
            )
        sequence.state = State.INTERRUPTED
        if stop_time == "now":
            stop_time = datetime.datetime.now(tz=datetime.timezone.utc)
        if not is_tz_aware(stop_time):
            raise ValueError("Stop time must be timezone aware")
        sequence.stop_time = stop_time.astimezone(datetime.timezone.utc).replace(
            tzinfo=None
        )
        return Success(None)

    def set_crashed(
        self,
        path: PureSequencePath,
        tb_summary: TracebackSummary,
        stop_time: datetime.datetime | Literal["now"],
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        sequence_result = _query_sequence_model(self._get_sql_session(), path)
        if is_failure(sequence_result):
            return sequence_result
        sequence = sequence_result.value
        if not State.is_transition_allowed(sequence.state, State.CRASHED):
            return Failure(
                InvalidStateTransitionError(
                    f"Sequence at {path} can't transition from {sequence.state} to "
                    f"{State.CRASHED}"
                )
            )
        sequence.state = State.CRASHED
        if stop_time == "now":
            stop_time = datetime.datetime.now(tz=datetime.timezone.utc)
        if not is_tz_aware(stop_time):
            raise ValueError("Stop time must be timezone aware")
        sequence.stop_time = stop_time.astimezone(datetime.timezone.utc).replace(
            tzinfo=None
        )
        set_exception_result = self._set_exception(path, tb_summary)
        assert not is_failure_type(set_exception_result, PathNotFoundError)
        assert not is_failure_type(set_exception_result, PathIsNotSequenceError)
        assert not is_failure_type(set_exception_result, SequenceNotCrashedError)
        assert_type(set_exception_result, Success[None])
        return Success(None)

    def reset_to_draft(
        self, path: PureSequencePath
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        return _reset_to_draft(self._get_sql_session(), path)

    def _set_device_configurations(
        self,
        path: PureSequencePath,
        device_configurations: Mapping[DeviceName, DeviceConfiguration],
    ) -> None:
        sequence = unwrap(self._query_sequence_model(path))
        if sequence.state != State.PREPARING:
            raise SequenceNotEditableError(path)
        sql_device_configs = []
        for name, device_configuration in device_configurations.items():
            type_name, content = self.serializer.dump_device_configuration(
                device_configuration
            )
            sql_device_configs.append(
                SQLDeviceConfiguration(
                    name=name, device_type=type_name, content=content
                )
            )
        sequence.device_configurations = sql_device_configs

    def get_device_configurations(
        self, path: PureSequencePath
    ) -> dict[DeviceName, DeviceConfiguration]:
        sequence = unwrap(self._query_sequence_model(path))
        if sequence.state == State.DRAFT:
            raise RuntimeError("Sequence has not been prepared yet")

        device_configurations = {}

        for device_configuration in sequence.device_configurations:
            constructed = self.serializer.load_device_configuration(
                device_configuration.device_type, device_configuration.content
            )
            device_configurations[device_configuration.name] = constructed
        return device_configurations

    def get_stats(
        self, path: PureSequencePath
    ) -> (
        Success[SequenceStats]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
    ):
        return _get_stats(self._get_sql_session(), path)

    def create_shot(
        self,
        shot_id: ShotId,
        shot_parameters: Mapping[DottedVariableName, Parameter],
        shot_data: Mapping[DataLabel, Data],
        shot_start_time: datetime.datetime,
        shot_end_time: datetime.datetime,
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotRunningError]
    ):
        return _create_shot(
            self._get_sql_session(),
            shot_id,
            shot_parameters,
            shot_data,
            shot_start_time,
            shot_end_time,
        )

    def get_shots(
        self, path: PureSequencePath
    ) -> (
        Success[list[ShotId]]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
    ):
        return _get_shots(self._get_sql_session(), path)

    def get_shot_parameters(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DottedVariableName, Parameter]:
        return _get_shot_parameters(self._get_sql_session(), path, shot_index)

    def get_all_shot_data(
        self, path: PureSequencePath, shot_index: int
    ) -> dict[DataLabel, Data]:
        return _get_all_shot_data(self._get_sql_session(), path, shot_index)

    def get_shot_data_by_label(self, data: DataId) -> Data:
        return _get_shot_data_by_label(self._get_sql_session(), data)

    def get_shot_data_by_labels(
        self, path: PureSequencePath, shot_index: int, data_labels: Set[DataLabel]
    ) -> Mapping[DataLabel, Data]:
        return _get_shot_data_by_labels(
            self._get_sql_session(), path, shot_index, data_labels
        )

    def get_shot_start_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        return _get_shot_start_time(self._get_sql_session(), path, shot_index)

    def get_shot_end_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        return _get_shot_end_time(self._get_sql_session(), path, shot_index)

    def get_sequences_in_state(self, state: State) -> Iterable[PureSequencePath]:
        stmt = (
            select(SQLSequencePath).join(SQLSequence).where(SQLSequence.state == state)
        )
        result = self._get_sql_session().execute(stmt).scalars().all()
        return (PureSequencePath(row.path) for row in result)

    def _query_path_model(
        self, path: PureSequencePath
    ) -> (
        Success[SQLSequencePath] | Failure[PathNotFoundError] | Failure[PathIsRootError]
    ):
        return _query_path_model(self._get_sql_session(), path)

    def _query_sequence_model(
        self, path: PureSequencePath
    ) -> (
        Success[SQLSequence]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
    ):
        return _query_sequence_model(self._get_sql_session(), path)

    def _query_shot_model(
        self, path: PureSequencePath, shot_index: int
    ) -> (
        Success[SQLShot]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[ShotNotFoundError]
    ):
        return _query_shot_model(self._get_sql_session(), path, shot_index)

    def _get_sql_session(self) -> sqlalchemy.orm.Session:
        # noinspection PyProtectedMember
        return self.parent_session._get_sql_session()


def _convert_from_unknown(value: int | Unknown) -> Optional[int]:
    if isinstance(value, Unknown):
        return None
    elif isinstance(value, int):
        return value
    else:
        assert_never(value)


def _convert_to_unknown(value: Optional[int]) -> int | Unknown:
    if value is None:
        return Unknown()
    elif isinstance(value, int):
        return value
    else:
        assert_never(value)


def _is_sequence(
    session: Session, path: PureSequencePath
) -> Success[bool] | Failure[PathNotFoundError]:
    path_model_result = _query_path_model(session, path)
    if isinstance(path_model_result, Failure):
        if is_failure_type(path_model_result, PathNotFoundError):
            return path_model_result
        else:
            assert_type(path_model_result, Failure[PathIsRootError])
            return Success(False)
    else:
        path_model = path_model_result.value
        return Success(bool(path_model.sequence))


def _get_exceptions(
    session: Session, path: PureSequencePath
) -> (
    Success[Optional[TracebackSummary]]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[SequenceNotCrashedError]
):
    sequence_model_query = _query_sequence_model(session, path)
    match sequence_model_query:
        case Success(sequence_model):
            assert isinstance(sequence_model, SQLSequence)
            if sequence_model.state != State.CRASHED:
                return Failure(SequenceNotCrashedError(path))
            exception_model = sequence_model.exception_traceback
            if exception_model is None:
                return Success(None)
            else:
                traceback_summary = cattrs.structure(
                    exception_model.content, TracebackSummary
                )
                return Success(traceback_summary)
        case Failure() as failure:
            return failure


def _set_exception(
    session: Session, path: PureSequencePath, exception: TracebackSummary
) -> (
    Success[None]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[SequenceNotCrashedError]
):
    sequence_model_query = _query_sequence_model(session, path)
    match sequence_model_query:
        case Success(sequence_model):
            assert isinstance(sequence_model, SQLSequence)
            if sequence_model.state != State.CRASHED:
                return Failure(SequenceNotCrashedError(path))
            if sequence_model.exception_traceback is not None:
                raise RuntimeError("Exception already set")
            content = cattrs.unstructure(exception, TracebackSummary)
            sequence_model.exception_traceback = SQLExceptionTraceback(content=content)
            return Success(None)
        case Failure() as failure:
            return failure


def _get_stats(
    session: Session, path: PureSequencePath
) -> (
    Success[SequenceStats]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
):
    result = _query_sequence_model(session, path)

    def extract_stats(sequence: SQLSequence) -> SequenceStats:
        number_shot_query = select(func.count()).select_from(
            select(SQLShot).where(SQLShot.sequence == sequence).subquery()
        )
        number_shot_run = session.execute(number_shot_query).scalar_one()
        return SequenceStats(
            state=sequence.state,
            start_time=(
                sequence.start_time.replace(tzinfo=datetime.timezone.utc)
                if sequence.start_time is not None
                else None
            ),
            stop_time=(
                sequence.stop_time.replace(tzinfo=datetime.timezone.utc)
                if sequence.stop_time is not None
                else None
            ),
            number_completed_shots=number_shot_run,
            expected_number_shots=_convert_to_unknown(
                sequence.expected_number_of_shots
            ),
        )

    return result.map(extract_stats)


def _get_sequence_global_parameters(
    session: Session, path: PureSequencePath
) -> (
    Success[ParameterNamespace]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[SequenceNotLaunchedError]
):
    sequence_result = _query_sequence_model(session, path)
    if is_failure(sequence_result):
        return sequence_result
    sequence = sequence_result.value

    if sequence.state == State.DRAFT:
        return Failure(
            SequenceNotLaunchedError(f"Sequence at {path} is in DRAFT state")
        )

    parameters_content = sequence.parameters.content

    return Success(
        serialization.converters["json"].structure(
            parameters_content, ParameterNamespace
        )
    )


def _get_iteration_configuration(
    session: Session, sequence: PureSequencePath, serializer: SerializerProtocol
) -> IterationConfiguration:
    sequence_model = unwrap(_query_sequence_model(session, sequence))
    return serializer.construct_sequence_iteration(
        sequence_model.iteration.content,
    )


def _set_iteration_configuration(
    session: Session,
    sequence: PureSequencePath,
    iteration_configuration: IterationConfiguration,
    serializer: SerializerProtocol,
) -> (
    None
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[SequenceNotEditableError]
):
    sequence_model_result = _query_sequence_model(session, sequence)
    if is_failure(sequence_model_result):
        return sequence_model_result
    sequence_model = sequence_model_result.content()
    if not sequence_model.state.is_editable():
        raise SequenceNotEditableError(sequence)
    iteration_content = serializer.dump_sequence_iteration(iteration_configuration)
    sequence_model.iteration.content = iteration_content
    expected_number_shots = iteration_configuration.expected_number_shots()
    sequence_model.expected_number_of_shots = _convert_from_unknown(
        expected_number_shots
    )


def _get_time_lanes(
    session: Session, sequence_path: PureSequencePath, serializer: SerializerProtocol
) -> TimeLanes | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]:
    sequence_model_result = _query_sequence_model(session, sequence_path)
    if is_failure(sequence_model_result):
        return sequence_model_result
    sequence_model = sequence_model_result.content()
    return serializer.structure_time_lanes(sequence_model.time_lanes.content)


def _set_time_lanes(
    session: Session,
    sequence_path: PureSequencePath,
    time_lanes: TimeLanes,
    serializer: SerializerProtocol,
) -> (
    None
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[SequenceNotEditableError]
):
    sequence_model_result = _query_sequence_model(session, sequence_path)
    if is_failure(sequence_model_result):
        return sequence_model_result
    sequence_model = sequence_model_result.content()
    if not sequence_model.state.is_editable():
        return Failure(SequenceNotEditableError(sequence_path))
    sequence_model.time_lanes.content = serializer.unstructure_time_lanes(time_lanes)


def _get_shots(
    session: Session, path: PureSequencePath
) -> (
    Success[list[ShotId]] | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]
):
    sql_sequence = _query_sequence_model(session, path)

    def extract_shots(sql_sequence: SQLSequence) -> list[ShotId]:
        return [ShotId(path, shot.index) for shot in sql_sequence.shots]

    return sql_sequence.map(extract_shots)


def _get_shot_parameters(
    session: Session, path: PureSequencePath, shot_index: int
) -> Mapping[DottedVariableName, Parameter]:
    stmt = (
        select(SQLShotParameter.content)
        .join(SQLShot)
        .where(SQLShot.index == shot_index)
        .join(SQLSequence)
        .join(SQLSequencePath)
        .where(SQLSequencePath.path == str(path))
    )

    result = session.execute(stmt).scalar_one_or_none()
    if result is not None:
        return serialization.converters["json"].structure(
            result, dict[DottedVariableName, bool | int | float | Quantity]
        )
    # This will raise the proper error if the shot was not found.
    unwrap(_query_shot_model(session, path, shot_index))
    raise AssertionError("Unreachable code")


def _create_shot(
    session: Session,
    shot_id: ShotId,
    shot_parameters: Mapping[DottedVariableName, Parameter],
    shot_data: Mapping[DataLabel, Data],
    shot_start_time: datetime.datetime,
    shot_end_time: datetime.datetime,
) -> (
    Success[None]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[SequenceNotRunningError]
):
    sequence_result = _query_sequence_model(session, shot_id.sequence_path)
    if is_failure(sequence_result):
        return sequence_result
    sequence = sequence_result.value
    if sequence.state != State.RUNNING:
        return Failure(SequenceNotRunningError(shot_id.sequence_path))
    if shot_id.index < 0:
        raise ValueError("Shot index must be non-negative")
    if sequence.expected_number_of_shots is not None:
        if shot_id.index >= sequence.expected_number_of_shots:
            raise ValueError(
                f"Shot index must be less than the expected number of shots "
                f"({sequence.expected_number_of_shots})"
            )

    parameters = serialize_shot_parameters(shot_parameters)

    array_data, structured_data = serialize_data(shot_data)

    shot = SQLShot(
        sequence=sequence,
        index=shot_id.index,
        parameters=SQLShotParameter(content=parameters),
        array_data=array_data,
        structured_data=structured_data,
        start_time=shot_start_time.astimezone(datetime.timezone.utc).replace(
            tzinfo=None
        ),
        end_time=shot_end_time.astimezone(datetime.timezone.utc).replace(tzinfo=None),
    )
    session.add(shot)
    return Success(None)


def _get_all_shot_data(
    session: Session, path: PureSequencePath, shot_index: int
) -> dict[DataLabel, Data]:
    shot_model = unwrap(_query_shot_model(session, path, shot_index))
    arrays = shot_model.array_data
    structured_data = shot_model.structured_data
    result = {}
    for array in arrays:
        result[array.label] = np.frombuffer(array.bytes_, dtype=array.dtype).reshape(
            array.shape
        )
    for data in structured_data:
        result[data.label] = data.content
    return result


def _get_shot_data_by_label(
    session: Session,
    data: DataId,
) -> Data:
    return _get_shot_data_by_labels(
        session, data.shot_id.sequence_path, data.shot_id.index, {data.data_label}
    )[data.data_label]


def _get_shot_data_by_labels(
    session: Session,
    path: PureSequencePath,
    shot_index: int,
    data_labels: Set[DataLabel],
) -> dict[DataLabel, Data]:
    content = unwrap(_query_data_model(session, path, shot_index, data_labels))

    data = {}

    for label, value in content.items():
        if isinstance(value, SQLStructuredShotData):
            data[label] = value.content
        elif isinstance(value, SQLShotArray):
            data[label] = np.frombuffer(value.bytes_, dtype=value.dtype).reshape(
                value.shape
            )
        else:
            assert_never(value)
    return data


def _get_shot_start_time(
    session: Session, path: PureSequencePath, shot_index: int
) -> datetime.datetime:
    shot_model = unwrap(_query_shot_model(session, path, shot_index))
    return shot_model.start_time.replace(tzinfo=datetime.timezone.utc)


def _get_shot_end_time(
    session: Session, path: PureSequencePath, shot_index: int
) -> datetime.datetime:
    shot_model = unwrap(_query_shot_model(session, path, shot_index))
    return shot_model.end_time.replace(tzinfo=datetime.timezone.utc)


def _query_data_model(
    session: Session,
    path: PureSequencePath,
    shot_index: int,
    data_labels: Set[DataLabel],
) -> (
    Success[dict[DataLabel, SQLShotArray | SQLStructuredShotData]]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[ShotNotFoundError]
    | Failure[DataNotFoundError]
):
    data = {}
    data_labels = set(data_labels)
    stmt = (
        select(SQLStructuredShotData)
        .where(SQLStructuredShotData.label.in_(data_labels))
        .join(SQLShot)
        .where(SQLShot.index == shot_index)
        .join(SQLSequence)
        .join(SQLSequencePath)
        .where(SQLSequencePath.path == str(path))
    )
    results = session.execute(stmt).all()
    for (result,) in results:
        data[result.label] = result
        data_labels.remove(result.label)
    if not data_labels:
        return Success(data)
    stmt = (
        select(SQLShotArray)
        .where(SQLShotArray.label.in_(data_labels))
        .join(SQLShot)
        .where(SQLShot.index == shot_index)
        .join(SQLSequence)
        .join(SQLSequencePath)
        .where(SQLSequencePath.path == str(path))
    )
    results = session.execute(stmt).all()
    for (result,) in results:
        data[result.label] = result
        data_labels.remove(result.label)
    if not data_labels:
        return Success(data)
    shot_result = _query_shot_model(session, path, shot_index)
    match shot_result:
        case Success():
            return Failure(DataNotFoundError(data_labels))
        case Failure() as failure:
            return failure


def _query_sequence_model(
    session: Session, path: PureSequencePath
) -> (
    Success[SQLSequence] | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]
):
    stmt = (
        select(SQLSequence)
        .join(SQLSequencePath)
        .where(SQLSequencePath.path == str(path))
    )
    result = session.execute(stmt).scalar_one_or_none()
    if result is not None:
        return Success(result)
    else:
        # If we are not is the happy path, we need to check the reason why to be able to
        # return the correct error.
        path_result = _query_path_model(session, path)
        if isinstance(path_result, Success):
            return Failure(PathIsNotSequenceError(path))
        else:
            if is_failure_type(path_result, PathNotFoundError):
                return path_result
            else:
                assert_type(path_result, Failure[PathIsRootError])
                return Failure(PathIsNotSequenceError(path))


def _query_shot_model(
    session: Session, path: PureSequencePath, shot_index: int
) -> (
    Success[SQLShot]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[ShotNotFoundError]
):
    stmt = (
        select(SQLShot)
        .where(SQLShot.index == shot_index)
        .join(SQLSequence)
        .join(SQLSequencePath)
        .where(SQLSequencePath.path == str(path))
    )

    result = session.execute(stmt).scalar_one_or_none()
    if result is not None:
        return Success(result)
    else:
        # This function is fast for the happy path were the shot exists, but if it was
        # not found, we need to check the reason why to be able to return the correct
        # error.
        sequence_model_result = _query_sequence_model(session, path)
        match sequence_model_result:
            case Success():
                return Failure(
                    ShotNotFoundError(
                        f"Shot {shot_index} not found for sequence {path}"
                    )
                )
            case Failure() as failure:
                return failure
            case _:
                assert_never(sequence_model_result)


def _reset_to_draft(
    session: Session, path: PureSequencePath
) -> (
    Success[None]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[InvalidStateTransitionError]
):
    sequence_result = _query_sequence_model(session, path)
    if is_failure(sequence_result):
        return sequence_result
    sequence = sequence_result.value
    if not State.is_transition_allowed(sequence.state, State.DRAFT):
        return Failure(
            InvalidStateTransitionError(
                f"Sequence at {path} can't transition from {sequence.state} to "
                f"{State.DRAFT}"
            )
        )
    sequence.state = State.DRAFT
    sequence.start_time = None
    sequence.stop_time = None
    sequence.parameters.content = None
    if sequence.exception_traceback:
        session.delete(sequence.exception_traceback)
        sequence.exception_traceback = None
    delete_device_configurations = sqlalchemy.delete(SQLDeviceConfiguration).where(
        SQLDeviceConfiguration.sequence == sequence
    )
    session.execute(delete_device_configurations)

    delete_shots = sqlalchemy.delete(SQLShot).where(SQLShot.sequence == sequence)
    session.execute(delete_shots)
    return Success(None)


def serialize_data(
    data: Mapping[DataLabel, Data]
) -> tuple[list[SQLShotArray], list[SQLStructuredShotData]]:
    arrays = []
    structured_data = []
    for label, value in data.items():
        if not is_data(value):
            raise TypeError(f"Invalid data type for {label}: {type(value)}")
        if isinstance(value, np.ndarray):
            arrays.append(
                SQLShotArray(
                    label=label,
                    dtype=str(value.dtype),
                    shape=value.shape,
                    bytes_=value.tobytes(),
                )
            )
        else:
            structured_data.append(SQLStructuredShotData(label=label, content=value))
    return arrays, structured_data


def serialize_shot_parameters(
    shot_parameters: Mapping[DottedVariableName, Parameter]
) -> dict[str, serialization.JSON]:
    return {
        str(variable_name): serialization.converters["json"].unstructure(
            parameter, Parameter
        )
        for variable_name, parameter in shot_parameters.items()
    }


def is_tz_aware(dt: datetime.datetime) -> bool:
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None



from ._implementation import TimeLaneSerializer, LaneLoader, LaneDumper
from ._protocol import TimeLaneSerializerProtocol

__all__ = [
    "TimeLaneSerializerProtocol",
    "TimeLaneSerializer",
    "LaneLoader",
    "LaneDumper",
]



from collections.abc import Set
from typing import Optional, Literal

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QComboBox, QWidget

from caqtus.device.sequencer.trigger import (
    Trigger,
    SoftwareTrigger,
    ExternalClockOnChange,
    ExternalClock,
    ExternalTriggerStart,
)

AvailableTriggers = Literal[
    "Software", "External start", "External clock", "External adaptive clock"
]

DEFAULT_AVAILABLE_TRIGGERS: Set[AvailableTriggers] = {
    "Software",
    "External start",
    "External clock",
    "External adaptive clock",
}


class TriggerSelector(QComboBox):
    """A widget to select a trigger for a sequencer.

    Args:
        available_triggers: The triggers that can be selected.

            If a device does not support all trigger types, the triggers proposed by
            the editor can be limited by providing a subset of the available triggers.

            The default is all available triggers.

        parent: The parent widget.

    Signals:
        trigger_changed: Emitted when the trigger is changed, either by the user or
            programmatically.
    """

    trigger_changed = Signal()

    def __init__(
        self,
        available_triggers: Set[AvailableTriggers] = DEFAULT_AVAILABLE_TRIGGERS,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)
        proposed_triggers = sorted(available_triggers)
        self.tags_to_index = {tag: index for index, tag in enumerate(proposed_triggers)}
        self.index_to_tag = {index: tag for index, tag in enumerate(proposed_triggers)}
        self.addItems(proposed_triggers)
        self.currentIndexChanged.connect(self._on_trigger_changed)

    def set_trigger(self, trigger: Trigger) -> None:
        """Set the trigger to be displayed."""

        trigger_to_tag = {
            SoftwareTrigger: "Software",
            ExternalTriggerStart: "External start",
            ExternalClock: "External clock",
            ExternalClockOnChange: "External adaptive clock",
        }

        try:
            trigger_tag = trigger_to_tag[type(trigger)]
        except KeyError:
            raise ValueError(f"Unsupported trigger: {trigger}")

        try:
            trigger_index = self.tags_to_index[trigger_tag]
        except KeyError:
            raise ValueError(f"Unsupported trigger: {trigger}")

        self.setCurrentIndex(trigger_index)

    def get_trigger(self) -> Trigger:
        """Get the trigger currently selected."""

        index = self.currentIndex()

        trigger_tag = self.index_to_tag[index]

        tag_to_trigger = {
            "Software": SoftwareTrigger,
            "External start": ExternalTriggerStart,
            "External clock": ExternalClock,
            "External adaptive clock": ExternalClockOnChange,
        }

        return tag_to_trigger[trigger_tag]()

    def _on_trigger_changed(self, *args, **kwargs) -> None:
        self.trigger_changed.emit()



import abc
import bisect
import itertools
from collections.abc import MutableSequence, Iterable, Sequence, Iterator
from typing import TypeVar, Self, NewType, overload, Never

import attrs
from typing_extensions import deprecated

from caqtus.types.expression import Expression
from caqtus.utils.asserts import assert_length_changed

#: Represents the index of a single step in a time lane.
Step = NewType("Step", int)

#: Represents the index of a block of steps in a time lane.
Block = NewType("Block", int)

#: Represents the number of steps spanned by a block.
Span = NewType("Span", int)


@attrs.define(init=False, eq=False, repr=False)
class TimeLane[T](MutableSequence[T], abc.ABC):
    """Represents a sequence of values covering some steps in time.

    A time lane is a sequence of values that are associated with some time steps.
    The time steps are not explicitly stored in the lane.
    Instead, for a given lane, lane[i] is the value of the lane at the i-th step.

    Consecutive identical values in a lane can be grouped in a single block spanning
    multiple steps.

    TimeLane is an abstract class, generic over the type of values it contains.
    It provides common methods such as value access, insertion, and deletion.
    """

    # spanned_values[i] is the value of i-th block of the lane and the number of steps
    # it spans.
    _spanned_values: list[tuple[T, Span]] = attrs.field()

    # _bounds[i] is the step at which the i-th block starts (inclusive)
    # _bounds[i+1] is the step at which the i-th block ends (exclusive)
    _bounds: list[Step] = attrs.field(init=False, repr=False)

    @_spanned_values.validator  # type: ignore
    def validate_spanned_values(self, _, value):
        if not all(span >= 1 for _, span in value):
            raise ValueError("Span must be at least 1")

    def __init__(self, values: Iterable[T]):
        """Initialize the lane with the given values.

        This constructor will group consecutive values that share the same id into
        blocks.

        This means that the following three lanes have the same blocks with length 3, 2,
        and 1:
        DigitalTimeLane([(True, 3), (False, 2), (True, 1)])
        DigitalTimeLane([True, True, True, False, False, True])
        DigitalTimeLane([True] * 3 + [False] * 2 + [True])

        Note however that the two following lanes are equivalent:
        AnalogTimeLane([Expression("...")] * 2 + [Expression("...")] * 3)
        AnalogTimeLane([(Expression("..."), 2), (Expression("..."), 3)])
        but are different from:
        AnalogTimeLane([Expression("...")] * 5)

        """

        values_list = list(values)
        spanned_values = []
        for _, group in itertools.groupby(values_list, key=id):
            g = list(group)
            spanned_values.append((g[0], Span(len(g))))
        self._spanned_values = spanned_values
        self._bounds = compute_bounds(span for _, span in self._spanned_values)

    @classmethod
    def from_spanned_values(cls, spanned_values: Iterable[tuple[T, Span]]) -> Self:
        obj = cls.__new__(cls)
        obj._spanned_values = list(spanned_values)
        obj._bounds = compute_bounds(span for _, span in obj._spanned_values)
        return obj

    def spanned_values(self) -> Sequence[tuple[T, Span]]:
        """Returns the spanned values of the lane.

        For each block in the lane, returns a tuple of two elements: the value of the
        block and the number of steps spanned by the block.

        Examples:
            >>> from caqtus.types.timelane import DigitalTimeLane
            >>> lane = DigitalTimeLane([True, True, False, True])
            >>> lane.spanned_values()
            [(True, 2), (False, 1), (True, 1)]
        """

        return tuple(self._spanned_values)

    @property
    def number_blocks(self) -> int:
        """Returns the number of blocks in the lane."""

        return len(self._spanned_values)

    def get_bounds(self, step: Step) -> tuple[Step, Step]:
        """Returns the bounds of the block containing the given step."""

        step = self._normalize_step(step)
        if not (0 <= step < len(self)):
            raise IndexError(f"Index out of bounds: {step}")
        return self.get_block_bounds(find_containing_block(self._bounds, step))

    def block_values(self) -> Iterator[T]:
        """Returns an iterator over the block values.

        The length of the iterator is the number of blocks in the lane, not the number
        of steps.
        """

        return (value for value, _ in self._spanned_values)

    @deprecated("use block_values instead")
    def values(self) -> Iterator[T]:
        return self.block_values()

    def block_bounds(self) -> Iterator[tuple[Step, Step]]:
        """Iterates over the bounds of the blocks.

        Returns:
            An iterator over the bounds of the blocks.

            Its elements are tuples of two elements: the step (inclusive) at which the
            block starts and the step (exclusive) at which the block ends.

            The length of the iterator is the number of blocks in the lane.
        """

        return zip(self._bounds[:-1], self._bounds[1:], strict=True)

    def get_block_bounds(self, block: Block) -> tuple[Step, Step]:
        """Returns the bounds of the given block.

        Returns:
            A tuple of two elements: the step (inclusive) at which the block starts and
            the step (exclusive) at which the block ends.

        Raises:
            IndexError: If the block index is out of bounds.
        """

        if not (0 <= block < self.number_blocks):
            raise IndexError(f"Block index out of bounds: {block}")

        return self._bounds[block], self._bounds[block + 1]

    @deprecated("use block_bounds instead")
    def bounds(self) -> Iterator[tuple[Step, Step]]:
        return self.block_bounds()

    def _get_containing_block(self, index: Step) -> Block:
        return find_containing_block(self._bounds, index)

    def _get_block_span(self, block: Block) -> Span:
        return self._spanned_values[block][1]

    def get_block_value(self, block: Block) -> T:
        """Returns the value of the given block."""

        return self._spanned_values[block][0]

    def __len__(self):
        return self._bounds[-1]

    @overload
    def __getitem__(self, item: int) -> T: ...

    @overload
    def __getitem__(self, item: slice) -> Never: ...

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_value_at_step(Step(item))
        else:
            raise TypeError(f"Invalid type for item: {type(item)}")

    def _get_value_at_step(self, step: Step) -> T:
        step = self._normalize_step(step)
        if not (0 <= step < len(self)):
            raise IndexError(f"Step out of bounds: {step}")
        return self.get_block_value(find_containing_block(self._bounds, step))

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.set_value_at_step(Step(key), value)
        elif isinstance(key, slice):
            self.set_value_for_slice(key, value)
        else:
            raise TypeError(f"Invalid type for item: {type(key)}")

    def set_value_at_step(self, step: Step, value: T):
        step = self._normalize_step(step)
        if not (0 <= step < len(self)):
            raise IndexError(f"Step out of bounds: {step}")
        block = find_containing_block(self._bounds, step)
        start, stop = self.get_block_bounds(block)
        before_length = Span(step - start)
        after_length = Span(stop - step - 1)
        previous_value = self.get_block_value(block)
        insert_index = block
        if before_length > 0:
            self._spanned_values.insert(insert_index, (previous_value, before_length))
            insert_index += 1
        self._spanned_values[insert_index] = (value, Span(1))
        insert_index += 1
        if after_length > 0:
            self._spanned_values.insert(insert_index, (previous_value, after_length))
        self._bounds = compute_bounds(span for _, span in self._spanned_values)

    def set_value_for_slice(self, slice_: slice, value: T):
        start = self._normalize_step(
            Step(slice_.start if slice_.start is not None else 0)
        )
        stop = self._normalize_step(
            Step(slice_.stop if slice_.stop is not None else len(self))
        )
        if not (0 <= start <= stop <= len(self)):
            raise IndexError(f"Slice out of bounds: {slice_}")
        if slice_.step is not None:
            raise ValueError(f"Slice step must be None: {slice_}")
        before_block = find_containing_block(self._bounds, start)
        before_length = Span(start - self.get_block_bounds(before_block)[0])
        before_value = self.get_block_value(before_block)
        if stop == len(self):
            after_block = Block(len(self._spanned_values) - 1)
        else:
            after_block = find_containing_block(self._bounds, stop)
        after_length = Span(self.get_block_bounds(after_block)[1] - stop)
        after_value = self.get_block_value(after_block)
        del self._spanned_values[before_block : after_block + 1]
        insert_index = before_block
        if before_length > 0:
            self._spanned_values.insert(before_block, (before_value, before_length))
            insert_index += 1
        self._spanned_values.insert(insert_index, (value, Span(stop - start)))
        insert_index += 1
        if after_length > 0:
            self._spanned_values.insert(insert_index, (after_value, after_length))
        self._bounds = compute_bounds(span for _, span in self._spanned_values)

    def _normalize_step(self, step: Step) -> Step:
        if step < 0:
            step = Step(len(self) + step)
        return step

    def __delitem__(self, key):
        if isinstance(key, int):
            self.delete_step(Step(key))
        else:
            raise TypeError(f"Invalid type for item: {type(key)}")

    def delete_step(self, step: Step) -> None:
        """Delete a single step from the lane.

        The length of the lane is always exactly one less after this operation.
        """

        previous_length = len(self)

        step = self._normalize_step(step)
        if not (0 <= step < len(self)):
            raise IndexError(f"Index out of bounds: {step}")
        block = find_containing_block(self._bounds, step)
        if self._get_block_span(block) == 1:
            del self._spanned_values[block]
        else:
            self._spanned_values[block] = (
                self.get_block_value(block),
                Span(self._get_block_span(block) - 1),
            )
        self._bounds = compute_bounds(span for _, span in self._spanned_values)
        assert len(self) == previous_length - 1

    @assert_length_changed(+1)
    def insert(self, index: int, value: T) -> None:
        step = Step(index)
        step = self._normalize_step(step)
        if step == len(self):
            self._spanned_values.append((value, Span(1)))
            self._bounds.append(Step(self._bounds[-1] + 1))
            return
        if not (0 <= step < len(self)):
            raise IndexError(f"Step out of bounds: {step}")
        block = find_containing_block(self._bounds, step)
        start, stop = self.get_block_bounds(block)
        before_length = Span(step - start)
        after_length = Span(stop - step)
        previous_value = self.get_block_value(block)
        insert_index = block
        if before_length > 0:
            self._spanned_values.insert(insert_index, (previous_value, before_length))
            insert_index += 1
        self._spanned_values[insert_index] = (value, Span(1))
        insert_index += 1
        if after_length > 0:
            self._spanned_values.insert(insert_index, (previous_value, after_length))
        self._bounds = compute_bounds(span for _, span in self._spanned_values)

    def __repr__(self):
        to_concatenate = []
        for length, group in itertools.groupby(
            self._spanned_values, key=lambda x: x[1]
        ):
            if length == 1:
                to_concatenate.append(
                    f"[{', '.join(repr(value) for value, _ in group)}]"
                )
            else:
                to_concatenate.extend(
                    [f"[{repr(value)}] * {length}" for value, _ in group]
                )
        formatted = " + ".join(to_concatenate)
        return f"{type(self).__name__}({formatted})"

    def __eq__(self, other):
        if isinstance(other, TimeLane):
            return self._spanned_values == other._spanned_values
        elif isinstance(other, Sequence):
            if len(self) != len(other):
                return False
            return all(a == b for a, b in zip(self, other, strict=True))
        else:
            return NotImplemented


TimeLaneType = TypeVar("TimeLaneType", bound=TimeLane)


def compute_bounds(spans: Iterable[Span]) -> list[Step]:
    return [0] + list(itertools.accumulate(spans))  # type: ignore


def find_containing_block(bounds: Sequence[Step], index: Step) -> Block:
    return Block(bisect.bisect(bounds, index) - 1)


@attrs.define(eq=False)
class TimeLanes:
    """A collection of time lanes.

    Groups together multiple time lanes and the associated step names and durations.
    """

    step_names: list[str] = attrs.field(
        factory=list,
        validator=attrs.validators.deep_iterable(
            iterable_validator=attrs.validators.instance_of(list),
            member_validator=attrs.validators.instance_of(str),
        ),
        on_setattr=attrs.setters.validate,
    )
    step_durations: list[Expression] = attrs.field(
        factory=list,
        validator=attrs.validators.deep_iterable(
            iterable_validator=attrs.validators.instance_of(list),
            member_validator=attrs.validators.instance_of(Expression),
        ),
        on_setattr=attrs.setters.validate,
    )
    lanes: dict[str, TimeLane] = attrs.field(
        factory=dict,
        validator=attrs.validators.deep_mapping(
            key_validator=attrs.validators.instance_of(str),
            value_validator=attrs.validators.instance_of(TimeLane),
        ),
        on_setattr=attrs.setters.validate,
    )

    @classmethod
    def empty(cls) -> Self:
        return cls([], [], {})

    @step_names.validator  # type: ignore
    def validate_step_names(self, _, value):
        if not all(isinstance(v, str) for v in value):
            raise ValueError("All step names must be instances of str")

    @step_durations.validator  # type: ignore
    def validate_step_durations(self, _, value):
        if len(value) != len(self.step_names):
            raise ValueError(
                "The number of step durations must match the number of step names"
            )
        if not all(isinstance(v, Expression) for v in value):
            raise ValueError("All step durations must be instances of Expression")

    @lanes.validator  # type: ignore
    def validate_lanes(self, _, value):
        if not all(isinstance(v, TimeLane) for v in value.values()):
            raise ValueError("All lanes must be instances of TimeLane")
        for name, lane in value.items():
            if len(lane) != self.number_steps:
                raise ValueError(
                    f"Lane '{name}' does not have the same length as the time steps"
                )

    @property
    def number_steps(self) -> int:
        """The number of steps in the time lanes.

        The number of steps is the same as the length of each lane.
        """

        return len(self.step_names)

    @property
    def number_lanes(self) -> int:
        """Returns the number of lanes."""

        return len(self.lanes)

    def __setitem__(self, name: str, lane: TimeLane):
        """Sets the value of a lane.

        Raises:
            TypeError: If the lane value is not an instance of TimeLane.
            TypeError: If the lane name is not a string.
            ValueError: If the lane value has a different length than the other lanes.
        """

        if not isinstance(lane, TimeLane):
            raise TypeError(f"Invalid type for value: {type(lane)}")
        if not isinstance(name, str):
            raise TypeError(f"Invalid type for key: {type(name)}")
        if len(lane) != self.number_steps:
            raise ValueError("All lanes must have the same length")

        self.lanes[name] = lane

    def __getitem__(self, key: str) -> TimeLane:
        """Returns the value of a lane.

        Raises:
            KeyError: If the lane name is not found.
        """

        return self.lanes[key]

    def __eq__(self, other):
        if not isinstance(other, TimeLanes):
            return NotImplemented
        return (
            self.step_names == other.step_names
            and self.step_durations == other.step_durations
            and self.lanes == other.lanes
            # We say that two TimeLanes are different if the order of the lanes is
            # different.
            # This is because the order of the lanes is important for the user.
            and list(self.lanes.keys()) == list(other.lanes.keys())
        )



"""Provides an implementation of experiment sessions using SQL databases."""

from ._session_maker import (
    PostgreSQLExperimentSessionMaker,
    PostgreSQLStorageManager,
    PostgreSQLConfig,
    SQLiteStorageManager,
    SQLiteConfig,
    SQLStorageManager,
)

__all__ = [
    "PostgreSQLExperimentSessionMaker",
    "PostgreSQLConfig",
    "SQLiteConfig",
    "SQLiteStorageManager",
    "PostgreSQLStorageManager",
    "SQLStorageManager",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'settings_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_SettingsDialog(object):
    def setupUi(self, SettingsDialog):
        if not SettingsDialog.objectName():
            SettingsDialog.setObjectName("SettingsDialog")
        SettingsDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(SettingsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.xAxisLabel = QLabel(SettingsDialog)
        self.xAxisLabel.setObjectName("xAxisLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.xAxisLabel)

        self.x_line_edit = QLineEdit(SettingsDialog)
        self.x_line_edit.setObjectName("x_line_edit")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.x_line_edit)

        self.yAxisLabel = QLabel(SettingsDialog)
        self.yAxisLabel.setObjectName("yAxisLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.yAxisLabel)

        self.y_line_edit = QLineEdit(SettingsDialog)
        self.y_line_edit.setObjectName("y_line_edit")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.y_line_edit)

        self.hueLabel = QLabel(SettingsDialog)
        self.hueLabel.setObjectName("hueLabel")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.hueLabel)

        self.hue_line_edit = QLineEdit(SettingsDialog)
        self.hue_line_edit.setObjectName("hue_line_edit")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.hue_line_edit)

        self.verticalLayout.addLayout(self.formLayout)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )

        self.verticalLayout.addItem(self.verticalSpacer)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.ok_button = QPushButton(SettingsDialog)
        self.ok_button.setObjectName("ok_button")

        self.horizontalLayout.addWidget(self.ok_button)

        self.horizontalLayout.setStretch(0, 1)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(SettingsDialog)

        QMetaObject.connectSlotsByName(SettingsDialog)

    # setupUi

    def retranslateUi(self, SettingsDialog):
        SettingsDialog.setWindowTitle(
            QCoreApplication.translate("SettingsDialog", "Dialog", None)
        )
        self.xAxisLabel.setText(
            QCoreApplication.translate("SettingsDialog", "x axis", None)
        )
        self.yAxisLabel.setText(
            QCoreApplication.translate("SettingsDialog", "y axis", None)
        )
        self.hueLabel.setText(QCoreApplication.translate("SettingsDialog", "hue", None))
        self.hue_line_edit.setPlaceholderText(
            QCoreApplication.translate("SettingsDialog", "None", None)
        )
        self.ok_button.setText(QCoreApplication.translate("SettingsDialog", "Ok", None))

    # retranslateUi



import functools
import logging

from typing import Callable, TypeVar, ParamSpec

_P = ParamSpec("_P")
_T = TypeVar("_T")


def log_call(
    logger: logging.Logger, log_args=True, log_result=True, level: int = logging.DEBUG
):
    """Decorator to log calls to a function.

    Args:
        logger: The logger to use to record the call.
        log_args: Indicate if the arguments passed to the function needs to be logged.
        log_result: Indicate if the result of the function needs to be logged.
        level: The logging level to use.
    """

    def decorator(func: Callable[_P, _T]) -> Callable[_P, _T]:
        @functools.wraps(func)
        def wrapper(
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> _T:
            if log_args:
                logger.log(
                    level,
                    "Calling %s with args %r and kwargs %r.",
                    func.__name__,
                    args,
                    kwargs,
                )
            else:
                logger.log(level, "Calling %s.", func.__name__)
            result = func(*args, **kwargs)
            if log_result:
                logger.log(
                    level, "Finished calling %s, returning %s.", func.__name__, result
                )
            else:
                logger.log(level, "Finished calling %s.", func.__name__)
            return result

        return wrapper

    return decorator



import functools
from collections.abc import Callable
from typing import Any

import cattrs.strategies

from caqtus.device.output_transform import (
    EvaluableOutput,
    Transformation,
    LinearInterpolation,
)
from caqtus.types.expression import Expression
from caqtus.utils.serialization import copy_converter

_converter = copy_converter()


@_converter.register_structure_hook
def structure_evaluable_output(data, _) -> EvaluableOutput:
    if isinstance(data, str):
        return Expression(data)
    else:
        return _converter.structure(data, Transformation)


# We need to register subclasses once they have been imported and defined.
cattrs.strategies.include_subclasses(
    Transformation,
    converter=_converter,
    subclasses=(LinearInterpolation,),
    union_strategy=functools.partial(
        cattrs.strategies.configure_tagged_union, tag_name="type"
    ),
)


def get_converter():
    return _converter


def get_structure_hook(type_: Any) -> Callable[[Any], Any]:
    hook = _converter.get_structure_hook(type_)

    def structure_hook(data):
        return hook(data, type_)

    return structure_hook


def get_unstructure_hook(type_: Any) -> Callable[[Any], Any]:
    hook = _converter.get_unstructure_hook(type_)
    return hook



from __future__ import annotations

import abc
import copy
from typing import Optional, Any, TypeVar

import attrs
from PySide6.QtCore import (
    QObject,
    QModelIndex,
    QAbstractListModel,
    Qt,
    QSize,
    QPersistentModelIndex,
)
from PySide6.QtGui import QAction, QFont, QUndoCommand, QUndoStack
from PySide6.QtWidgets import QMenu

import caqtus.gui.qtutil.qabc as qabc
from caqtus.types.timelane import TimeLane, Step

T = TypeVar("T")
L = TypeVar("L", bound=TimeLane)

_DEFAULT_INDEX = QModelIndex()


class TimeLaneModel[L: TimeLane](QAbstractListModel, metaclass=qabc.QABCMeta):
    """An abstract list model to represent a time lane.

    This class inherits from :class:`PySide6.QtCore.QAbstractListModel` and can be
    used to represent a lane in the timelanes editor.

    It is meant to be subclassed for each lane type that needs to be represented in
    the timelanes editor.
    Some common methods are implemented here, but subclasses will need to implement at
    least the abstract methods: :meth:`data`, :meth:`setData`, :meth:`insertRow`.
    In addition, subclasses may want to override :meth:`flags` to change the item flags
    for the cells in the lane.
    The :meth:`get_cell_context_actions` method can be overridden to add context menu
    actions to the cells in the lane.

    Warning:
        All user edits to the data model should be done through the undo stack of
        the model.

    """

    # Ignore some lint rules for this class as PySide6 models have a lot of camelCase
    # methods.
    # ruff: noqa: N802

    def __init__(
        self,
        name: str,
        lane: L,
        parent: Optional[QObject] = None,
    ):
        super().__init__(parent)
        self.__name = name
        self.__lane: L = lane
        self.__undo_stack = QUndoStack(self)

    def set_undo_stack(self, undo_stack: QUndoStack) -> None:
        """Set the undo stack for the model."""

        self.__undo_stack = undo_stack

    def name(self) -> str:
        """Return the name of the lane represented by this model."""

        return self.__name

    def get_lane(self) -> L:
        """Return a copy of the lane represented by this model."""

        return copy.deepcopy(self.__lane)

    def set_lane[
        L_contra: TimeLane
    ](self: TimeLaneModel[L_contra], lane: L_contra) -> None:
        """Set the lane represented by this model.

        This method does not push changes to the undo stack.
        """

        self.beginResetModel()
        self.__lane = copy.deepcopy(lane)
        self.endResetModel()

    def rowCount(
        self, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> int:
        """Return the number of steps in the lane."""

        return len(self.__lane)

    @abc.abstractmethod
    def data(
        self,
        index: QModelIndex | QPersistentModelIndex,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        """Return the data to be shown to the user for the given index and role.

        See :meth:`PySide6.QtCore.QAbstractItemModel.data` for more information on the
        roles that can be used.

        This method must be implemented by subclasses, typically by calling the
        :meth:`lane_value` method and returning the appropriate value for the given
        role.
        """

        return None

    def lane_value[T](self: TimeLaneModel[TimeLane[T]], step: int) -> T:
        """Return the value of the underlying lane at the given step."""

        return self.__lane[step]

    @abc.abstractmethod
    def setData(self, index, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        """Set the data for the given index and role.

        See :meth:`PySide6.QtCore.QAbstractItemModel.setData` for more information on
        the roles that can be used.

        This method must be implemented by subclasses, typically by calling the
        :meth:`set_lane_value` method with an adequate value.
        """

        raise NotImplementedError

    def set_lane_value[
        T
    ](self: TimeLaneModel[TimeLane[T]], step: int, value: T) -> bool:
        """Set the value of the underlying lane at the given step.

        The value of the whole block the step is part of is set to the new value.

        The change is pushed to the undo stack.

        Returns:
            True if the value was set successfully, False otherwise.
        """

        if not (0 <= step < len(self.__lane)):
            return False
        self.__undo_stack.push(
            self.SetValueCommand(
                model=self,
                step=step,
                previous_value=self.lane_value(step),
                new_value=value,
            )
        )
        return True

    @attrs.define(slots=False, kw_only=True)
    class SetValueCommand(QUndoCommand):
        """An undo/redo command to set the value of a step in a time lane model."""

        model: TimeLaneModel
        step: int
        previous_value: Any
        new_value: Any

        def __attrs_post_init__(self):
            super().__init__(
                f"change value for Step {self.step} of {self.model.name()} from "
                f"<{self.previous_value}> to <{self.new_value}>"
            )

        def redo(self):
            self.model._set_value_without_undo(self.step, self.new_value)

        def undo(self):
            self.model._set_value_without_undo(self.step, self.previous_value)

    def _set_value_without_undo[
        T
    ](self: TimeLaneModel[TimeLane[T]], step: int, value: T) -> None:
        """Set the value of the underlying lane without pushing an undo command.

        The step must be a valid index for the lane.
        The value of the whole block the step is part of is set to the new value.
        """

        assert 0 <= step < len(self.__lane)
        start, stop = self.__lane.get_bounds(Step(step))
        self.__lane[start:stop] = value
        self.dataChanged.emit(self.index(step), self.index(step))

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        """Return the data for the model header."""

        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self.name()
            elif orientation == Qt.Orientation.Vertical:
                return f"Step {section}"
        elif role == Qt.ItemDataRole.FontRole:
            font = QFont()
            font.setBold(True)
            return font

    def flags(self, index) -> Qt.ItemFlag:
        """Return the flags for the given index.

        By default, the flags are set to `ItemIsEnabled`, `ItemIsEditable`, and
        `ItemIsSelectable`.
        """

        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsSelectable
        )

    @abc.abstractmethod
    def insertRow(
        self, row: int, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        """Insert a row at the given row index.

        This method must be implemented by subclasses, typically by calling the
        :meth:`insert_lane_value` method with an adequate value.
        """

        raise NotImplementedError

    def insert_lane_value[
        T
    ](self: TimeLaneModel[TimeLane[T]], step: int, value: T) -> bool:
        """Insert a value at the given row index.

        The action is pushed to the undo stack.
        """

        if not (0 <= step <= len(self.__lane)):
            return False
        self.__undo_stack.push(
            self.InsertValueCommand(model=self, step=step, new_value=value)
        )
        return True

    @attrs.define(slots=False)
    class InsertValueCommand(QUndoCommand):
        model: TimeLaneModel
        step: int
        new_value: Any

        def __attrs_post_init__(self):
            super().__init__(f"insert step {self.step}")

        def redo(self):
            self.model._insert_lane_value_without_undo(self.step, self.new_value)

        def undo(self):
            self.model._remove_step_without_undo(self.step)

    def _insert_lane_value_without_undo[
        T
    ](self: TimeLaneModel[TimeLane[T]], step: int, value: T) -> None:
        """Insert a value at the given row index without pushing an undo command.

        The step must be a valid index for the lane.
        """

        assert 0 <= step <= len(self.__lane)
        self.beginInsertRows(QModelIndex(), step, step)
        if step == len(self.__lane):
            self.__lane.append(value)
        else:
            start, stop = self.__lane.get_bounds(Step(step))
            self.__lane.insert(step, value)
            if start < step < stop:
                self.__lane[start : stop + 1] = self.__lane[start]
        self.endInsertRows()

    def removeRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        """Remove a row at the given row index.

        The action is pushed to the undo stack.
        """

        if not (0 <= row < len(self.__lane)):
            return False

        start, stop = self.__lane.get_bounds(Step(row))
        self.__undo_stack.push(
            self.RemoveStepCommand(
                model=self, step=row, value=self.lane_value(row), start=start, stop=stop
            )
        )
        return True

    @attrs.define(slots=False)
    class RemoveStepCommand(QUndoCommand):
        model: TimeLaneModel
        step: int
        value: Any
        start: int
        stop: int

        def __attrs_post_init__(self):
            super().__init__(f"remove step {self.step}")

        def redo(self):
            self.model._remove_step_without_undo(self.step)

        def undo(self):
            self.model._reinsert_step_without_undo(
                self.step, self.value, self.start, self.stop
            )

    def _remove_step_without_undo(self, step: int) -> None:
        """Remove a step at the given row index without pushing an undo command.

        The step must be a valid index for the lane.
        """

        assert 0 <= step < len(self.__lane)
        self.beginRemoveRows(QModelIndex(), step, step)
        del self.__lane[step]
        self.endRemoveRows()

    def _reinsert_step_without_undo(
        self, step: int, value, start: int, stop: int
    ) -> None:
        assert 0 <= step <= len(self.__lane)
        self.beginInsertRows(QModelIndex(), step, step)
        self._insert_lane_value_without_undo(step, value)
        self._expand_step_without_undo(step, start, stop - 1)
        self.endInsertRows()

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        break_span_action = QAction("Break block")
        break_span_action.triggered.connect(lambda: self.break_span(index))
        return [break_span_action]

    def span(self, index) -> QSize:
        """Return the span of the cell at the given index."""

        start, stop = self.__lane.get_bounds(Step(index.row()))
        if index.row() == start:
            return QSize(1, stop - start)
        else:
            return QSize(1, 1)

    def break_span(self, index: QModelIndex) -> bool:
        """Break the block of the cell at the given index.

        The action is pushed to the undo stack.
        """

        start, stop = self.__lane.get_bounds(Step(index.row()))
        self.__undo_stack.push(
            self.BreakSpanCommand(
                model=self,
                step=index.row(),
                start=start,
                stop=stop,
                lane=self.get_lane(),
            )
        )
        return True

    @attrs.define(slots=False)
    class BreakSpanCommand(QUndoCommand):
        model: TimeLaneModel
        step: int
        start: int
        stop: int
        lane: TimeLane

        def __attrs_post_init__(self):
            super().__init__(f"break span at step {self.step}")

        def redo(self):
            self.model._break_span_without_undo(self.step, self.start, self.stop)

        def undo(self):
            self.model.set_lane(self.lane)

    def _break_span_without_undo(self, step: int, start: int, stop: int) -> None:
        value = self.__lane[step]
        for i in range(start, stop):
            self.__lane[i] = value
        self.dataChanged.emit(self.index(start), self.index(stop - 1))

    def expand_step(self, step: int, start: int, stop: int) -> bool:
        """Expand the step at the given index to the given range.

        The action is pushed to the undo stack.
        """

        if not (0 <= step < len(self.__lane)):
            return False
        self.__undo_stack.push(
            self.ExpandStepCommand(
                model=self, step=step, start=start, stop=stop, lane=self.get_lane()
            )
        )
        return True

    @attrs.define(slots=False)
    class ExpandStepCommand(QUndoCommand):
        model: TimeLaneModel
        step: int
        start: int
        stop: int
        lane: TimeLane

        def __attrs_post_init__(self):
            super().__init__(f"expand Step {self.step}")

        def redo(self):
            self.model._expand_step_without_undo(self.step, self.start, self.stop)

        def undo(self):
            self.model.set_lane(self.lane)

    def _expand_step_without_undo(self, step: int, start: int, stop: int) -> None:
        value = self.__lane[step]
        self.__lane[start : stop + 1] = value
        self.dataChanged.emit(self.index(start), self.index(stop - 1))

    def get_header_context_actions(self) -> list[QAction | QMenu]:
        """Return a list of context menu actions for the lane header."""

        return []

    def simplify(self) -> None:
        """Simplify the lane by merging contiguous blocks of the same value.

        The action is pushed to the undo stack.
        """

        self.__undo_stack.push(
            self.SimplifyCommand(model=self, lane=copy.deepcopy(self.__lane))
        )

    @attrs.define(slots=False)
    class SimplifyCommand(QUndoCommand):
        model: TimeLaneModel
        lane: TimeLane

        def __attrs_post_init__(self):
            super().__init__("simplify lane")

        def redo(self):
            self.model._simplify_without_undo()

        def undo(self):
            self.model.set_lane(self.lane)

    def _simplify_without_undo(self) -> None:
        self.beginResetModel()
        start = 0
        for i in range(1, len(self.__lane)):
            if self.__lane[i] != self.__lane[start]:
                self.__lane[start:i] = self.__lane[start]
                start = i
        self.__lane[start:] = self.__lane[start]
        self.endResetModel()



from __future__ import annotations

import functools
from typing import Mapping, Any, SupportsFloat

import attrs
import numpy as np

import caqtus.formatter as fmt
from caqtus.shot_compilation import ShotContext
from caqtus.shot_compilation.lane_compilation import DimensionedSeries
from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Pattern,
    Concatenated,
    concatenate,
    Repeated,
)
from caqtus.types.expression import Expression
from caqtus.types.recoverable_exceptions import InvalidTypeError, InvalidValueError
from caqtus.types.units import Unit, Quantity, InvalidDimensionalityError
from caqtus.types.variable_name import DottedVariableName
from ..channel_output import ChannelOutput
from ...timing import TimeStep, ns


@attrs.define
class BroadenLeft(ChannelOutput):
    """Indicates that output should go high before the input pulses go high.

    The output y(t) of this operation should be high when any of the input x(s) is high
    for s in [t, t + width].

    The operation is only valid for boolean inputs, and it will produce a boolean
    output.

    It is meant to be used to compensate for finite rise times in the hardware.
    For example, if a shutter takes 10 ms to open, and we want to open it at time t, we
    can use this operation to start opening the shutter at time t - 10 ms.
    """

    input_: ChannelOutput = attrs.field(
        validator=attrs.validators.instance_of(ChannelOutput),
        on_setattr=attrs.setters.validate,
    )
    width: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ):
        instruction = self.input_.evaluate(
            required_time_step, prepend, append, shot_context
        )
        if instruction.values.dtype != np.bool_:
            raise InvalidTypeError("Can't broaden non boolean instruction")
        width = self.width.evaluate(shot_context.get_parameters())
        if not isinstance(width, Quantity):
            raise InvalidTypeError(
                f"Width {fmt.expression(self.width)} does not evaluate to a quantity, "
                f"got {fmt.type_(type(width))}"
            )
        if not width.is_compatible_with(Unit("s")):
            raise InvalidDimensionalityError(
                f"Width {fmt.expression(self.width)} does not have units of time, got "
                f"{fmt.unit(width.units)}"
            )
        seconds = width.to(Unit("s")).magnitude
        if seconds < 0:
            raise InvalidValueError(
                f"Width {fmt.expression(self.width)} evaluates to a negative value"
            )
        ticks = duration_to_ticks(seconds, required_time_step)

        broadened, bleed = _broaden_left(instruction.values, ticks)

        return DimensionedSeries(broadened, instruction.units)

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        return self.input_.evaluate_max_advance_and_delay(time_step, variables)


@functools.singledispatch
def _broaden_left(instruction, width: int) -> tuple[TimedInstruction[np.bool_], int]:
    """Broaden the instruction to the left by n steps.

    Returns:
        - result: The expanded instruction, where the high values are expanded to the
            left, i.e. result[i] = any(instruction[i:i+width+1])
        - bleed: The number of steps before this instruction that must be set to True,
            i.e. bleed = max(0, width - first) where first is the index of the first
            high value in the instruction if it exists, otherwise bleed = 0.
    """

    raise NotImplementedError(
        f"Don't know how to broaden instruction of type {type(instruction)}"
    )


@_broaden_left.register
def _expand_pattern_left(instruction: Pattern, width: int):
    pulse_length = min(len(instruction), width + 1)
    pulse = np.full(pulse_length, True)
    convolution = np.convolve(instruction.array, pulse)
    result = convolution[pulse_length - 1 :]
    high_indices = instruction.array.nonzero()[0]
    if len(high_indices) == 0:
        excess = 0
    else:
        first_high_index = int(high_indices[0])  # need to avoid numpy integers
        excess = max(0, width - first_high_index)
    return Pattern.create_without_copy(result), excess


@_broaden_left.register
def _expand_concatenated_left(instruction: Concatenated, width: int):
    new_instructions = []
    bleed = 0
    for sub_instruction in reversed(instruction.instructions):
        expanded, new_bleed = _broaden_left(sub_instruction, width)
        overwritten_length = min(bleed, len(expanded))
        overwritten = Pattern([True]) * overwritten_length
        new_instructions.append(overwritten)
        kept = expanded[: len(expanded) - len(overwritten)]
        new_instructions.append(kept)
        bleed -= len(expanded)
        bleed = max(new_bleed, bleed)
    return concatenate(*reversed(new_instructions)), bleed


@_broaden_left.register
def _expand_repeated_left(repeated: Repeated, width: int):
    expanded, bleed = _broaden_left(repeated.instruction, width)
    if bleed == 0:
        # This is a special were expanding the instruction has no effect on the previous
        # instructions.
        return expanded * repeated.repetitions, bleed
    if bleed >= len(expanded):
        # This is a special case where the previous instructions are completely
        # overwritten by the expanded instruction, so we can return a simplified
        # instruction.
        instr = Pattern([True]) * len(expanded) * (repeated.repetitions - 1) + expanded
        return instr, bleed
    overwritten_length = min(bleed, len(expanded))
    overwritten = Pattern([True]) * overwritten_length
    kept = expanded[: len(expanded) - len(overwritten)]
    left_instr = kept + overwritten
    if expanded == left_instr:
        return expanded * repeated.repetitions, bleed
    else:
        instr = (kept + overwritten) * (repeated.repetitions - 1) + expanded
        return instr, bleed


def duration_to_ticks(duration: SupportsFloat, time_step: TimeStep) -> int:
    """Returns the nearest number of ticks for a given duration and time step.

    Args:
        duration: The duration in seconds.
        time_step: The time step in nanoseconds.
    """

    dt = time_step * ns

    rounded = round(float(duration) / float(dt))
    if not isinstance(rounded, int):
        raise TypeError(f"Expected integer number of ticks, got {rounded}")
    return rounded



from __future__ import annotations

import abc
import contextlib
from collections.abc import Callable, AsyncGenerator
from collections.abc import Mapping
from typing import Protocol

from caqtus.device import DeviceController
from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.device.remote import DeviceProxy
from caqtus.shot_compilation import SequenceContext
from caqtus.types.data import DataLabel, Data
from ._initialize_devices import create_devices
from ._shot_compiler import ShotCompilerProtocol
from ._shot_event_dispatcher import DeviceRunConfig, ShotEventDispatcher
from ._shot_primitives import DeviceParameters
from ..device_manager_extension import DeviceManagerExtensionProtocol


class ShotRunnerProtocol(Protocol):
    """Interface for running a shot."""

    @abc.abstractmethod
    async def run_shot(
        self, shot_parameters: DeviceParameters
    ) -> Mapping[DataLabel, Data]:
        """Run a shot.

        Args:
            shot_parameters: The parameters for the shot.
        """

        ...


type ShotRunnerFactory = Callable[
    [SequenceContext, ShotCompilerProtocol, DeviceManagerExtensionProtocol],
    contextlib.AbstractAsyncContextManager[ShotRunnerProtocol],
]
"""A function called to create a object that can run a shot."""


class ShotRunner(ShotRunnerProtocol):
    def __init__(
        self,
        devices: Mapping[DeviceName, DeviceProxy],
        controller_types: Mapping[DeviceName, type[DeviceController]],
    ):
        if set(devices.keys()) != set(controller_types.keys()):
            raise ValueError("The devices and controller_types must have the same keys")
        self.devices = devices
        self.controller_types = controller_types

    async def run_shot(
        self,
        shot_parameters: DeviceParameters,
    ) -> Mapping[DataLabel, Data]:
        event_dispatcher = ShotEventDispatcher(
            {
                name: DeviceRunConfig(
                    device=self.devices[name],
                    controller_type=self.controller_types[name],
                    parameters=shot_parameters.device_parameters[name],
                )
                for name in self.devices
            }
        )
        return await event_dispatcher.run_shot(shot_parameters.timeout)


@contextlib.asynccontextmanager
async def create_shot_runner(
    sequence_context: SequenceContext,
    shot_compiler: ShotCompilerProtocol,
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> AsyncGenerator[ShotRunner, None]:
    """Creates and acquires resources for running a shot.

    Returns:
        A context manager that yields a shot runner.
    """

    initialization_parameters = shot_compiler.compile_initialization_parameters()

    device_configurations_in_use = {
        name: sequence_context.get_device_configuration(name)
        for name in initialization_parameters
    }

    device_types = {
        name: device_manager_extension.get_device_type(config)
        for name, config in device_configurations_in_use.items()
    }

    async with create_devices(
        initialization_parameters=initialization_parameters,
        device_configs=device_configurations_in_use,
        device_types=device_types,
        device_manager_extension=device_manager_extension,
    ) as devices_in_use:
        shot_runner = _create_shot_runner(
            device_proxies=devices_in_use,
            device_configurations=device_configurations_in_use,
            device_manager_extension=device_manager_extension,
        )

        yield shot_runner


def _create_shot_runner(
    device_proxies: Mapping[DeviceName, DeviceProxy],
    device_configurations: Mapping[DeviceName, DeviceConfiguration],
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> ShotRunner:
    device_controller_types = {
        name: device_manager_extension.get_device_controller_type(
            device_configurations[name]
        )
        for name in device_proxies
    }
    shot_runner = ShotRunner(device_proxies, device_controller_types)
    return shot_runner



from collections.abc import Sequence
from typing import assert_never

import numpy as np

from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Pattern,
    concatenate,
)
from caqtus.types.expression import Expression
from caqtus.types.parameter import Parameters
from caqtus.types.timelane import DigitalTimeLane
from .._evaluation import evaluate_time_dependent_digital_expression
from ..timing import Time, number_ticks


def compile_digital_lane(
    lane: DigitalTimeLane,
    step_start_times: Sequence[Time],
    time_step: Time,
    parameters: Parameters,
) -> TimedInstruction[np.bool]:
    """Compile a digital lane into a sequence of instructions.

    Args:
        lane: The digital lane to compile.
        step_start_times: The start times of each step.
            The length of this sequence must be equal to the number of steps in the
            lane plus one, with the last element being the total duration.
        time_step: The time step for discretizing the time, in seconds.
        parameters: The parameters to use when evaluating expressions in the lane.
    """

    if len(lane) != len(step_start_times) - 1:
        raise ValueError(
            f"Number of steps in lane ({len(lane)}) does not match number of step "
            f"start times ({len(step_start_times) - 1})"
        )

    instructions = []
    for cell_value, (start, stop) in zip(
        lane.block_values(), lane.block_bounds(), strict=True
    ):
        if isinstance(cell_value, bool):
            length = number_ticks(
                step_start_times[start], step_start_times[stop], time_step
            )
            instructions.append(Pattern([cell_value]) * length)
        elif isinstance(cell_value, Expression):
            instr = evaluate_time_dependent_digital_expression(
                cell_value,
                parameters,
                step_start_times[start],
                step_start_times[stop],
                time_step,
            )
            instructions.append(instr)
        else:
            assert_never(cell_value)
    return concatenate(*instructions)



from collections.abc import Iterable

import attrs
import numpy as np

from ._roi import ROI


@attrs.define
class RectangularROI(ROI):
    """Rectangular region of interest inside an image.

    Attributes:
        x: Horizontal coordinate of the left column of the roi, in pixels.
        width: Number of columns in the roi.
        y: Vertical coordinate of the bottom row of the roi, in pixels.
        height: Number of rows in the roi.
    """

    x: int = attrs.field(
        validator=[attrs.validators.instance_of(int), attrs.validators.ge(0)],
        on_setattr=attrs.setters.validate,
    )
    width: int = attrs.field(
        validator=[attrs.validators.instance_of(int), attrs.validators.ge(0)],
        on_setattr=attrs.setters.validate,
    )
    y: int = attrs.field(
        validator=[attrs.validators.instance_of(int), attrs.validators.ge(0)],
        on_setattr=attrs.setters.validate,
    )
    height: int = attrs.field(
        validator=[attrs.validators.instance_of(int), attrs.validators.ge(0)],
        on_setattr=attrs.setters.validate,
    )

    @x.validator  # type: ignore
    def _validate_x(self, _, x):
        if x >= self.original_image_size[0]:
            raise ValueError("x must be smaller than original_width")

    @width.validator  # type: ignore
    def _validate_width(self, _, width):
        if self.x + width > self.original_image_size[0]:
            raise ValueError("x + width must be smaller than original_width")

    @y.validator  # type: ignore
    def _validate_y(self, _, y):
        if y >= self.original_image_size[1]:
            raise ValueError("y must be smaller than original_height")

    @height.validator  # type: ignore
    def _validate_height(self, _, height):
        if self.y + height > self.original_image_size[1]:
            raise ValueError("y + height must be smaller than original_height")

    def get_mask(self) -> np.ndarray:
        """A boolean array with the same shape as the original image.

        True values indicate that the pixel is part of the region of interest."""

        mask = np.full(self.original_image_size, False)
        mask[self.y : self.y + self.height, self.x : self.x + self.width] = True
        return mask

    def get_indices(self) -> tuple[Iterable[int], Iterable[int]]:
        """Return the pixels in the original image that are part of the ROI."""

        raise NotImplementedError

    @property
    def left(self) -> int:
        """Return the left column (included) of the roi."""

        return self.x

    @property
    def right(self) -> int:
        """Return the right column (included) of the roi."""

        return self.x + self.width - 1

    @property
    def bottom(self) -> int:
        """Return the bottom row (included) of the roi."""

        return self.y

    @property
    def top(self) -> int:
        """Return the top row (included) of the roi."""

        return self.y + self.height - 1



from __future__ import annotations

import functools
import typing
from collections.abc import Sequence

import attrs
from PySide6.QtWidgets import QWidget, QFormLayout, QLabel

from ._editor_builder import (
    EditorBuilder,
    EditorBuildingError,
    EditorFactory,
    TypeNotRegisteredError,
)
from .._int_editor import IntegerEditor
from .._string_editor import StringEditor
from .._value_editor import ValueEditor


def build_attrs_class_editor[
    T: attrs.AttrsInstance
](
    cls: type[T], builder: EditorBuilder, **attr_overrides: AttributeOverride
) -> EditorFactory[T]:
    """Build an editor for attrs class.

    This function will build a form editor with a list of widgets for the attributes of
    the class.

    The label for each widget is the name of the attribute, prettified by removing
    underscores and capitalizing the first letter of the first word.

    Args:
        cls: The attrs class to build the editor for.
        builder: The editor builder used to build editors for the class attributes.
        **attr_overrides: If a named argument corresponds to one of the
            attributes, the object passed as argument will override the automated ui
            generation for that attribute.
    """

    fields: tuple[attrs.Attribute, ...] = attrs.fields(cls)

    if any(isinstance(field.type, str) for field in fields):
        # PEP 563 annotations - need to be resolved.
        attrs.resolve_types(cls)

    attr_ui_infos = []
    for field in fields:
        override = attr_overrides.pop(field.name, None)
        try:
            label = get_attribute_label(field, override)
            editor_factory = get_attribute_editor_factory(field, override, builder)
            tooltip = get_attribute_tooltip(field, override)
            order = override.order if override is not None else 0
        except Exception as e:
            raise AttributeEditorBuildingError(cls, field) from e
        attr_ui_infos.append(
            AttributeUIInfo(
                field_name=field.name,
                label=label,
                editor_factory=editor_factory,
                tooltip=tooltip,
                order=order,
            )
        )
    if attr_overrides:
        raise ValueError(
            f"Unknown attribute overrides: {', '.join(attr_overrides.keys())}"
        )
    attr_ui_infos.sort(key=lambda ui_info: ui_info.order)

    return functools.partial(AttrsEditor, cls, tuple(attr_ui_infos))


class AttrsEditor[T: attrs.AttrsInstance](ValueEditor[T]):
    """A generic editor for attrs classes."""

    def __init__(self, cls: type[T], ui_specs: Sequence[AttributeUIInfo]) -> None:
        self._widget = QWidget()
        self._ui_specs = ui_specs
        self._cls = cls

        layout = QFormLayout()
        self._widget.setLayout(layout)
        for ui_spec in ui_specs:
            editor = ui_spec.editor_factory()
            setattr(self, ui_spec.editor_name, editor)
            label = QLabel(ui_spec.label)
            if ui_spec.tooltip is not None:
                label.setToolTip(ui_spec.tooltip)
            layout.addRow(label, editor.widget)

    @typing.override
    def set_value(self, value: T) -> None:
        for ui_spec in self._ui_specs:
            editor = getattr(self, ui_spec.editor_name)
            assert isinstance(editor, ValueEditor)
            editor.set_value(getattr(value, ui_spec.field_name))

    # TODO: Figure out why pyright report this method as an incompatible override
    @typing.override
    def read_value(self) -> T:  # type: ignore[reportIncompatibleMethodOverride]
        attribute_values = {}
        for ui_spec in self._ui_specs:
            editor = getattr(self, ui_spec.editor_name)
            assert isinstance(editor, ValueEditor)
            attribute_values[ui_spec.field_name] = editor.read_value()
        return self._cls(**attribute_values)

    @typing.override
    def set_editable(self, editable: bool) -> None:
        for ui_spec in self._ui_specs:
            editor = getattr(self, ui_spec.editor_name)
            assert isinstance(editor, ValueEditor)
            editor.set_editable(editable)

    @property
    @typing.override
    def widget(self) -> QWidget:
        return self._widget


@attrs.frozen
class AttributeOverride:
    """Overrides for the automatic generation of an attribute editor.

    Attributes:
        label: The label to use for the attribute.
        editor_factory: The editor factory to use for the attribute.
        tooltip: The tooltip to show for the attribute.
        order: The order in which to show the attribute in the editor.
            Attributes will be listed in ascending order.
            Default is 0.
    """

    label: str | None = None
    editor_factory: EditorFactory | None = None
    tooltip: str | None = None
    order: int = 0


@attrs.frozen
class AttributeUIInfo:
    field_name: str
    label: str
    editor_factory: EditorFactory
    tooltip: str | None
    order: int

    @property
    def editor_name(self) -> str:
        return attr_to_editor_name(self.field_name)


def get_attribute_label(
    attr: attrs.Attribute,
    override: AttributeOverride | None,
) -> str:
    label = None
    if override is not None:
        label = override.label
    if label is None:
        label = prettify_snake_case(attr.name)
    return label


def get_attribute_editor_factory(
    attr: attrs.Attribute,
    override: AttributeOverride | None,
    builder: EditorBuilder,
) -> EditorFactory:
    editor_factory = None
    if override is not None:
        editor_factory = override.editor_factory
    if editor_factory is None:
        if attr.type is None:
            raise ValueError("Attribute has no type annotation")
        try:
            editor_factory = builder.build_editor(attr.type)
        except TypeNotRegisteredError:
            if attr.type is int:
                editor_factory = IntegerEditor
            elif attr.type is str:
                editor_factory = StringEditor
            else:
                raise
    return editor_factory


def get_attribute_tooltip(
    attr: attrs.Attribute,
    override: AttributeOverride | None,
) -> str | None:
    tooltip = None
    if override is not None:
        tooltip = override.tooltip
    # TODO: Extract tooltip from documentation
    # Would need to parse the docstring of the class and find the attribute's docstring.
    # In addition, might need to parse the docstring of parent classes if the attribute
    # is inherited.
    return tooltip


def prettify_snake_case(name: str) -> str:
    if name:
        words = name.split("_")
        words[0] = words[0].title()
        return " ".join(words)
    else:
        return name


def attr_to_editor_name(name: str) -> str:
    return f"editor_{name}"


class AttributeEditorBuildingError(EditorBuildingError):
    def __init__(self, cls: type[attrs.AttrsInstance], attribute: attrs.Attribute):
        msg = f"Could not build editor for attribute '{attribute.name}' of {cls}"
        super().__init__(msg)



from __future__ import annotations

from collections.abc import Mapping, Callable
from typing import SupportsFloat, Protocol

import numpy

from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.types.units import DimensionalityError
from caqtus.types.variable_name import VariableName
from ._scalar import Scalar

SCALAR_FUNCTIONS: Mapping[VariableName, ScalarFunction]


class InvalidArgumentCountError(InvalidTypeError):
    pass


class ScalarFunction(Protocol):
    """A function that operates on scalars."""

    def __call__(self, *args: Scalar) -> Scalar:
        """Call the function with the given arguments.

        The function must accept a variable number of arguments, even if the underlying
        function is only defined for a specific number of arguments.
        This is because the user can mess up and enter the wrong number of arguments,
        but the function must still handle the error gracefully and raise the correct
        exception.

        Raises:
            InvalidArgumentCountError: if the number of arguments passed to the function
                is not correct.
            RecoverableError: if the function cannot be evaluated due to the user's
                input.
        """

        ...


def float_to_scalar_function(
    function: Callable[[float], SupportsFloat]
) -> ScalarFunction:
    """Convert a function operating on a float to a function operating on scalars."""

    def wrapper(*args: Scalar):
        if len(args) != 1:
            raise InvalidArgumentCountError(
                f"{function.__name__}() can only be called with one argument, got "
                f"{len(args)}."
            )
        value = args[0]
        try:
            converted = float(value)
        except (ValueError, DimensionalityError):
            raise InvalidTypeError(
                f"{function.__name__}() expected a number, got {value!r}."
            ) from None
        return float(function(converted))

    return wrapper


SCALAR_FUNCTIONS = {
    VariableName("abs"): float_to_scalar_function(abs),
    VariableName("arccos"): float_to_scalar_function(numpy.arccos),
    VariableName("arcsin"): float_to_scalar_function(numpy.arcsin),
    VariableName("arctan"): float_to_scalar_function(numpy.arctan),
    VariableName("cos"): float_to_scalar_function(numpy.cos),
    VariableName("sin"): float_to_scalar_function(numpy.sin),
    VariableName("tan"): float_to_scalar_function(numpy.tan),
    VariableName("exp"): float_to_scalar_function(numpy.exp),
    VariableName("log"): float_to_scalar_function(numpy.log),
    VariableName("sqrt"): float_to_scalar_function(numpy.sqrt),
}



import functools
import operator
from collections.abc import Sequence
from typing import assert_never, Any

from typing_extensions import TypeIs

import caqtus.formatter as fmt
import caqtus_parsing.nodes as nodes
from caqtus.types.expression import Expression
from caqtus.types.parameter import Parameters
from caqtus.types.recoverable_exceptions import EvaluationError, InvalidTypeError
from caqtus.types.units import Quantity, is_scalar_quantity, Unit, DimensionalityError
from caqtus_parsing import parse, InvalidSyntaxError
from ._constants import CONSTANTS
from ._exceptions import (
    UndefinedParameterError,
    InvalidOperationError,
    UndefinedUnitError,
)
from ._functions import SCALAR_FUNCTIONS
from ._scalar import Scalar
from ._units import units


def evaluate_scalar_expression(
    expression: Expression, parameters: Parameters
) -> Scalar:
    """Evaluate a scalar expression.

    Args:
        expression: The expression to evaluate.
        parameters: The parameters to use in the evaluation.

    Returns:
        The result of the evaluation.

    Raises:
        EvaluationError: if an error occurred during evaluation, with the reason for the
            error as the exception cause.
    """

    try:
        ast = parse(str(expression))
        return evaluate_expression(ast, parameters)
    except (EvaluationError, InvalidSyntaxError) as error:
        raise EvaluationError(
            f"Could not evaluate {fmt.expression(expression)}."
        ) from error


def evaluate_bool_expression(
    expression: nodes.Expression, parameters: Parameters
) -> bool:
    value = evaluate_expression(expression, parameters)
    if not isinstance(value, bool):
        raise InvalidTypeError(
            f"Expected {fmt.expression(expression)} to evaluate to a boolean, "
            f"but got {fmt.type_(type(value))}."
        )
    return value


def evaluate_float_expression(
    expression: nodes.Expression, parameters: Parameters
) -> float:
    value = evaluate_expression(expression, parameters)
    try:
        return float(value)
    except DimensionalityError as error:
        raise InvalidOperationError(
            f"Could not convert {fmt.expression(expression)} to a float."
        ) from error


def evaluate_expression(expression: nodes.Expression, parameters: Parameters) -> Scalar:
    match expression:
        case int() | float():
            return expression
        case nodes.Variable() as variable:
            return evaluate_scalar_variable(variable, parameters)
        case (
            nodes.Add()
            | nodes.Subtract()
            | nodes.Multiply()
            | nodes.Divide()
            | nodes.Power() as binary_operator
        ):
            return evaluate_binary_operator(binary_operator, parameters)
        case nodes.Plus() | nodes.Minus() as unary_operator:
            return evaluate_unary_operator(unary_operator, parameters)
        case nodes.Quantity():
            return evaluate_quantity(expression)
        case nodes.Call():
            return evaluate_function_call(expression, parameters)
        case _:  # pragma: no cover
            assert_never(expression)


def evaluate_scalar_variable(
    variable: nodes.Variable, parameters: Parameters
) -> Scalar:
    name = variable.name
    if name in parameters:
        # We can use str as key instead of DottedVariableName because they have the
        # same hash.
        return parameters[name]  # type: ignore[reportArgumentType]
    elif name in CONSTANTS:
        return CONSTANTS[name]
    else:
        raise UndefinedParameterError(f"Parameter {name} is not defined.")


def evaluate_function_call(
    function_call: nodes.Call,
    parameters: Parameters,
) -> Scalar:
    function_name = function_call.function
    try:
        # We can use str as key instead of VariableName because they have the
        # same hash.
        function = SCALAR_FUNCTIONS[function_name]  # type: ignore[reportArgumentType]
    except KeyError:
        raise UndefinedFunctionError(
            f"Function {function_name} is not defined."
        ) from None
    arguments = [
        evaluate_expression(argument, parameters) for argument in function_call.args
    ]
    return function(*arguments)


def evaluate_binary_operator(
    binary_operator: nodes.BinaryOperator,
    parameters: Parameters,
) -> Scalar:
    left = evaluate_expression(binary_operator.left, parameters)
    right = evaluate_expression(binary_operator.right, parameters)
    match binary_operator:
        case nodes.Add():
            result = left + right
        case nodes.Subtract():
            result = left - right
        case nodes.Multiply():
            result = left * right
        case nodes.Divide():
            result = left / right
        case nodes.Power(exponent):
            if not isinstance(right, (int, float)):
                raise InvalidOperationError(
                    f"The exponent {exponent} must be a real number, not {right}."
                )
            result = left**right
        case _:  # pragma: no cover
            assert_never(binary_operator)
    assert is_scalar(result)
    return result


def evaluate_unary_operator(
    unary_operator: nodes.UnaryOperator,
    parameters: Parameters,
) -> Scalar:
    operand = evaluate_expression(unary_operator.operand, parameters)
    match unary_operator:
        case nodes.Plus():
            result = operand
        case nodes.Minus():
            result = -operand
        case _:  # pragma: no cover
            assert_never(unary_operator)
    assert is_scalar(result)
    return result


def evaluate_quantity(quantity: nodes.Quantity) -> Quantity[float]:
    magnitude = quantity.magnitude

    multiplicative_units = evaluate_units(quantity.multiplicative_units)

    if not quantity.divisional_units:
        return Quantity(magnitude, multiplicative_units)
    else:
        divisive_units = evaluate_units(quantity.divisional_units)
        total_units = multiplicative_units / divisive_units
        assert isinstance(total_units, Unit)
        return Quantity(magnitude, total_units)


@functools.lru_cache
def evaluate_units(unit_nodes: Sequence[nodes.UnitTerm]) -> Unit:
    assert unit_nodes
    accumulated_units = []
    for unit_term in unit_nodes:
        try:
            base = units[unit_term.unit]
        except KeyError:
            raise UndefinedUnitError(f"Unit {unit_term.unit} is not defined.") from None
        exponent = unit_term.exponent or 1
        accumulated_units.append(base**exponent)
    return functools.reduce(operator.mul, accumulated_units)


def is_scalar(value: Any) -> TypeIs[Scalar]:
    return isinstance(value, (int, bool, float)) or is_scalar_quantity(value)


class UndefinedFunctionError(EvaluationError):
    pass



# Form implementation generated from reading ui file '.\graphplot\visualizer_creators_selector.ui'
#
# Created by: PyQt6 UI code generator 6.6.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtWidgets


class Ui_VisualizerCreatorSelector(object):
    def setupUi(self, VisualizerCreatorSelector):
        VisualizerCreatorSelector.setObjectName("VisualizerCreatorSelector")
        VisualizerCreatorSelector.resize(400, 300)
        VisualizerCreatorSelector.setLayoutDirection(
            QtCore.Qt.LayoutDirection.LeftToRight
        )
        self.verticalLayout = QtWidgets.QVBoxLayout(VisualizerCreatorSelector)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.dataLoaderLabel = QtWidgets.QLabel(parent=VisualizerCreatorSelector)
        self.dataLoaderLabel.setObjectName("dataLoaderLabel")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.dataLoaderLabel
        )
        self._visualizer_combo_box = QtWidgets.QComboBox(
            parent=VisualizerCreatorSelector
        )
        self._visualizer_combo_box.setObjectName("_visualizer_combo_box")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.ItemRole.FieldRole, self._visualizer_combo_box
        )
        self.verticalLayout.addLayout(self.formLayout)
        self._settings_group = QtWidgets.QGroupBox(parent=VisualizerCreatorSelector)
        self._settings_group.setObjectName("_settings_group")
        self.verticalLayout.addWidget(self._settings_group)
        spacerItem = QtWidgets.QSpacerItem(
            20,
            40,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        self.verticalLayout.addItem(spacerItem)
        self._apply_button = QtWidgets.QPushButton(parent=VisualizerCreatorSelector)
        self._apply_button.setObjectName("_apply_button")
        self.verticalLayout.addWidget(self._apply_button)

        self.retranslateUi(VisualizerCreatorSelector)
        QtCore.QMetaObject.connectSlotsByName(VisualizerCreatorSelector)

    def retranslateUi(self, VisualizerCreatorSelector):
        _translate = QtCore.QCoreApplication.translate
        VisualizerCreatorSelector.setWindowTitle(
            _translate("VisualizerCreatorSelector", "Form")
        )
        self.dataLoaderLabel.setText(
            _translate("VisualizerCreatorSelector", "DataView")
        )
        self._settings_group.setTitle(
            _translate("VisualizerCreatorSelector", "Settings")
        )
        self._apply_button.setText(_translate("VisualizerCreatorSelector", "Apply"))



from typing import Optional

from caqtus.gui._common.NodeGraphQt import BaseNode


class LaneNode(BaseNode):
    __identifier__ = "caqtus.sequencer_node.source"
    NODE_NAME = "Lane"

    def __init__(self):
        super().__init__()
        self.add_output("out", multi_output=False, display_name=False)
        self.default_port = self.add_input("absent?", multi_input=False)
        self.add_text_input(
            "Lane",
            text="...",
            placeholder_text="lane name",
            tooltip="The name of the lane that should be ouput by this node.",
        )

    def set_lane_name(self, lane_name: str) -> None:
        self.set_property("Lane", lane_name)

    def get_lane_name(self) -> str:
        return str(self.get_property("Lane"))

    def get_default_node(self) -> Optional[BaseNode]:
        input_nodes = self.connected_input_nodes()[self.default_port]
        if len(input_nodes) == 0:
            return None
        elif len(input_nodes) == 1:
            return input_nodes[0]
        else:
            assert False, "There can't be multiple nodes connected to the input"



#!/usr/bin/python
from .base_node import BaseNode
from ..qgraphics.node_circle import CircleNodeItem


class BaseNodeCircle(BaseNode):
    """
    `Implemented in` ``v0.5.2``

    The ``NodeGraphQt.BaseNodeCircle`` is pretty much the same class as the
    :class:`NodeGraphQt.BaseNode` except with a different design.

    .. inheritance-diagram:: NodeGraphQt.BaseNodeCircle

    .. image:: ../_images/node_circle.png
        :width: 250px

    example snippet:

    .. code-block:: python
        :linenos:

        from NodeGraphQt import BaseNodeCircle

        class ExampleNode(BaseNodeCircle):

            # unique node identifier domain.
            __identifier__ = 'io.jchanvfx.github'

            # initial default node name.
            NODE_NAME = 'My Node'

            def __init__(self):
                super(ExampleNode, self).__init__()

                # create an input port.
                self.add_input('in')

                # create an output port.
                self.add_output('out')
    """

    NODE_NAME = "Circle Node"

    def __init__(self, qgraphics_item=None):
        super(BaseNodeCircle, self).__init__(qgraphics_item or CircleNodeItem)



from __future__ import annotations

import traceback
from typing import Optional

import attrs


@attrs.frozen
class TracebackSummary:
    """Represents a summary of an exception and its traceback.

    Attributes:
        exc_type: The fully-qualified name of the exception type, like
            `builtins.RuntimeError`.
        exc_msg: The exception message.
        notes: A list of notes that were attached to the traceback.
        cause: The exception that caused this exception, if any.
        context: The exception that was being handled when this exception was raised,
            if any.
        exceptions: A list of exceptions that were caught by this exception if it is an
            instance of BaseExceptionGroup.
    """

    exc_type: str
    exc_msg: str
    notes: Optional[list[str]]
    cause: Optional[TracebackSummary]
    context: Optional[TracebackSummary]
    exceptions: Optional[list[TracebackSummary]]

    @classmethod
    def from_exception(cls, exc: BaseException) -> TracebackSummary:
        tb = traceback.TracebackException.from_exception(exc)

        cause = cls.from_exception(exc.__cause__) if exc.__cause__ else None
        context = cls.from_exception(exc.__context__) if exc.__context__ else None
        if isinstance(exc, BaseExceptionGroup):
            exceptions = [cls.from_exception(e) for e in exc.exceptions]
        else:
            exceptions = None
        notes = tb.__notes__  # pyright: ignore[reportAttributeAccessIssue]
        exc_type = type(exc)
        exc_msg = str(exc)

        return cls(
            cause=cause,
            context=context,
            exceptions=exceptions,
            notes=notes,
            exc_type=f"{exc_type.__module__}.{exc_type.__qualname__}",
            exc_msg=exc_msg,
        )

    def exc_cls(self) -> str:
        """Return the exception class name without the module."""

        return self.exc_type.split(".")[-1]



import string
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from caqtus.types.expression import Expression
    from caqtus.types.variable_name import DottedVariableName
    from caqtus.device import DeviceName
    from caqtus.types.units import UnitLike


class CaqtusFormatter(string.Formatter):
    """Custom formatter for caqtus.

    Instances of this class can be used to format strings with the following format
    specifiers:
    - :device: to format a string as a device name
    - :device server: to format a string as a remote server name

    It is useful to use this formatter to display objects in the caqtus package
    consistently.
    This allows parsing the string produced by the formatter to find the fields.

    Allowed format specifiers:

    * :step: to format a :data:`tuple[int, str]` as a step index and name.
    * :device: to format a :class:`str` as the name of a device.
    * :device server: to format a :class:`str` as the name of a remote server.
    * :expression: to format a :class:`caqtus.types.expressions.Expression` as the
    value of an expression.
    * :parameter assignment: to format a
    :data:`tuple[str, caqtus.types.expressions.Expression]` as the name and value of a
    parameter assignment.
    * :device parameter: to format a :data:`typle[str, Any]` as the name and value of
    a parameter for a device.

    Example:
    ```
    formatter = CaqtusFormatter()
    print(formatter.format("{:device}", "name"))

    # Output: device: 'name'
    ```
    """

    def format_field(self, value, format_spec):
        if format_spec == "step":
            step_index, step_name = value
            if not isinstance(step_index, int):
                raise ValueError("Step index must be an integer")
            if not isinstance(step_name, str):
                raise ValueError("Step name must be a string")
            value = f"step {step_index} ({step_name})"
        elif format_spec == "device":
            value = f"device '{value}'"
        elif format_spec == "device server":
            value = f"device server '{value}'"
        elif format_spec == "expression":
            value = f"expression '{value}'"
        elif format_spec == "parameter assignment":
            parameter_name, parameter_value = value
            value = f"parameter assignment '{parameter_name}' = '{parameter_value}'"
        elif format_spec == "type":
            value = f"type '{value.__name__}'"
        elif format_spec == "shot":
            if not isinstance(value, int):
                raise ValueError("Shot number must be an integer")
            value = f"shot {value}"
        elif format_spec == "device parameter":
            parameter_name, parameter_value = value
            value = f"device parameter '{parameter_name}' = '{parameter_value}'"
        return super().format(value, "")


caqtus_formatter = CaqtusFormatter()


def fmt_param_assign(name: "DottedVariableName", expression: "Expression") -> str:
    return f"parameter assignment '{name}' = '{expression}'"


def device(name: "DeviceName") -> str:
    return f"device '{name}'"


def device_param(name: str, value: str) -> str:
    return f"device parameter '{name}' = '{value}'"


def expression(expression) -> str:
    return f"expression '{expression}'"


def unit(value: "UnitLike") -> str:
    return f"unit '{value}'"


def shot_param(name: "DottedVariableName") -> str:
    return f"shot parameter '{name}'"


def type_(value: type | str) -> str:
    if isinstance(value, type):
        return f"type '{value.__name__}'"
    elif isinstance(value, str):
        return f"type '{value}'"
    else:
        raise ValueError("Invalid type value")


def lane(lane_name: str) -> str:
    return f"lane '{lane_name}'"


def fmt(s: str, *args, **kwargs):
    return caqtus_formatter.format(s, *args, **kwargs)



#!/usr/bin/python
from collections import OrderedDict

from PySide6 import QtGui, QtCore, QtWidgets

from ..constants import (
    ITEM_CACHE_MODE,
    ICON_NODE_BASE,
    LayoutDirectionEnum,
    NodeEnum,
    PortEnum,
    PortTypeEnum,
    Z_VAL_NODE,
)
from ..errors import NodeWidgetError
from ..qgraphics.node_abstract import AbstractNodeItem
from ..qgraphics.node_overlay_disabled import XDisabledItem
from ..qgraphics.node_text_item import NodeTextItem
from ..qgraphics.port import PortItem, CustomPortItem


class NodeItem(AbstractNodeItem):
    """
    Base Node item.

    Args:
        name (str): name displayed on the node.
        parent (QtWidgets.QGraphicsItem): parent item.
    """

    def __init__(self, name="node", parent=None):
        super(NodeItem, self).__init__(name, parent)
        pixmap = QtGui.QPixmap(ICON_NODE_BASE)
        if pixmap.size().height() > NodeEnum.ICON_SIZE.value:
            pixmap = pixmap.scaledToHeight(
                NodeEnum.ICON_SIZE.value, QtCore.Qt.SmoothTransformation
            )
        self._properties["icon"] = ICON_NODE_BASE
        self._icon_item = QtWidgets.QGraphicsPixmapItem(pixmap, self)
        self._icon_item.setTransformationMode(QtCore.Qt.SmoothTransformation)
        self._text_item = NodeTextItem(self.name, self)
        self._x_item = XDisabledItem(self, "DISABLED")
        self._input_items = OrderedDict()
        self._output_items = OrderedDict()
        self._widgets = OrderedDict()
        self._proxy_mode = False
        self._proxy_mode_threshold = 70

    def post_init(self, viewer, pos=None):
        """
        Called after node has been added into the scene.

        Args:
            viewer (NodeGraphQt.widgets.viewer.NodeViewer): main viewer
            pos (tuple): the cursor pos if node is called with tab search.
        """
        if self.layout_direction == LayoutDirectionEnum.VERTICAL.value:
            font = QtGui.QFont()
            font.setPointSize(15)
            self.text_item.setFont(font)

            # hide port text items for vertical layout.
            if self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
                for text_item in self._input_items.values():
                    text_item.setVisible(False)
                for text_item in self._output_items.values():
                    text_item.setVisible(False)

    def _paint_horizontal(self, painter, option, widget):
        painter.save()
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.NoBrush)

        # base background.
        margin = 1.0
        rect = self.boundingRect()
        rect = QtCore.QRectF(
            rect.left() + margin,
            rect.top() + margin,
            rect.width() - (margin * 2),
            rect.height() - (margin * 2),
        )

        radius = 4.0
        painter.setBrush(QtGui.QColor(*self.color))
        painter.drawRoundedRect(rect, radius, radius)

        # light overlay on background when selected.
        if self.selected:
            painter.setBrush(QtGui.QColor(*NodeEnum.SELECTED_COLOR.value))
            painter.drawRoundedRect(rect, radius, radius)

        # node name background.
        padding = 3.0, 2.0
        text_rect = self._text_item.boundingRect()
        text_rect = QtCore.QRectF(
            text_rect.x() + padding[0],
            rect.y() + padding[1],
            rect.width() - padding[0] - margin,
            text_rect.height() - (padding[1] * 2),
        )
        if self.selected:
            painter.setBrush(QtGui.QColor(*NodeEnum.SELECTED_COLOR.value))
        else:
            painter.setBrush(QtGui.QColor(0, 0, 0, 80))
        painter.drawRoundedRect(text_rect, 3.0, 3.0)

        # node border
        if self.selected:
            border_width = 1.2
            border_color = QtGui.QColor(*NodeEnum.SELECTED_BORDER_COLOR.value)
        else:
            border_width = 0.8
            border_color = QtGui.QColor(*self.border_color)

        border_rect = QtCore.QRectF(
            rect.left(), rect.top(), rect.width(), rect.height()
        )

        pen = QtGui.QPen(border_color, border_width)
        pen.setCosmetic(self.viewer().get_zoom() < 0.0)
        path = QtGui.QPainterPath()
        path.addRoundedRect(border_rect, radius, radius)
        painter.setBrush(QtCore.Qt.NoBrush)
        painter.setPen(pen)
        painter.drawPath(path)

        painter.restore()

    def _paint_vertical(self, painter, option, widget):
        painter.save()
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.NoBrush)

        # base background.
        margin = 1.0
        rect = self.boundingRect()
        rect = QtCore.QRectF(
            rect.left() + margin,
            rect.top() + margin,
            rect.width() - (margin * 2),
            rect.height() - (margin * 2),
        )

        radius = 4.0
        painter.setBrush(QtGui.QColor(*self.color))
        painter.drawRoundedRect(rect, radius, radius)

        # light overlay on background when selected.
        if self.selected:
            painter.setBrush(QtGui.QColor(*NodeEnum.SELECTED_COLOR.value))
            painter.drawRoundedRect(rect, radius, radius)

        # top & bottom edge background.
        padding = 2.0
        height = 10
        if self.selected:
            painter.setBrush(QtGui.QColor(*NodeEnum.SELECTED_COLOR.value))
        else:
            painter.setBrush(QtGui.QColor(0, 0, 0, 80))
        for y in [rect.y() + padding, rect.height() - height - 1]:
            edge_rect = QtCore.QRectF(
                rect.x() + padding, y, rect.width() - (padding * 2), height
            )
            painter.drawRoundedRect(edge_rect, 3.0, 3.0)

        # node border
        border_width = 0.8
        border_color = QtGui.QColor(*self.border_color)
        if self.selected:
            border_width = 1.2
            border_color = QtGui.QColor(*NodeEnum.SELECTED_BORDER_COLOR.value)
        border_rect = QtCore.QRectF(
            rect.left(), rect.top(), rect.width(), rect.height()
        )

        pen = QtGui.QPen(border_color, border_width)
        pen.setCosmetic(self.viewer().get_zoom() < 0.0)
        painter.setBrush(QtCore.Qt.NoBrush)
        painter.setPen(pen)
        painter.drawRoundedRect(border_rect, radius, radius)

        painter.restore()

    def paint(self, painter, option, widget):
        """
        Draws the node base not the ports.

        Args:
            painter (QtGui.QPainter): painter used for drawing the item.
            option (QtGui.QStyleOptionGraphicsItem):
                used to describe the parameters needed to draw.
            widget (QtWidgets.QWidget): not used.
        """
        self.auto_switch_mode()
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            self._paint_horizontal(painter, option, widget)
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            self._paint_vertical(painter, option, widget)
        else:
            raise RuntimeError("Node graph layout direction not valid!")

    def mousePressEvent(self, event):
        """
        Re-implemented to ignore event if LMB is over port collision area.

        Args:
            event (QtWidgets.QGraphicsSceneMouseEvent): mouse event.
        """
        if event.button() == QtCore.Qt.LeftButton:
            for p in self._input_items.keys():
                if p.hovered:
                    event.ignore()
                    return
            for p in self._output_items.keys():
                if p.hovered:
                    event.ignore()
                    return
        super(NodeItem, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """
        Re-implemented to ignore event if Alt modifier is pressed.

        Args:
            event (QtWidgets.QGraphicsSceneMouseEvent): mouse event.
        """
        if event.modifiers() == QtCore.Qt.AltModifier:
            event.ignore()
            return
        super(NodeItem, self).mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        """
        Re-implemented to emit "node_double_clicked" signal.

        Args:
            event (QtWidgets.QGraphicsSceneMouseEvent): mouse event.
        """
        if event.button() == QtCore.Qt.LeftButton:
            if not self.disabled:
                # enable text item edit mode.
                items = self.scene().items(event.scenePos())
                if self._text_item in items:
                    self._text_item.set_editable(True)
                    self._text_item.setFocus()
                    event.ignore()
                    return

            viewer = self.viewer()
            if viewer:
                viewer.node_double_clicked.emit(self.id)
        super(NodeItem, self).mouseDoubleClickEvent(event)

    def itemChange(self, change, value):
        """
        Re-implemented to update pipes on selection changed.

        Args:
            change:
            value:
        """
        if change == QtWidgets.QGraphicsItem.ItemSelectedChange and self.scene():
            self.reset_pipes()
            if value:
                self.highlight_pipes()
            self.setZValue(Z_VAL_NODE)
            if not self.selected:
                self.setZValue(Z_VAL_NODE + 1)

        return super(NodeItem, self).itemChange(change, value)

    def _tooltip_disable(self, state):
        """
        Updates the node tooltip when the node is enabled/disabled.

        Args:
            state (bool): node disable state.
        """
        tooltip = "<b>{}</b>".format(self.name)
        if state:
            tooltip += ' <font color="red"><b>(DISABLED)</b></font>'
        tooltip += "<br/>{}<br/>".format(self.type_)
        self.setToolTip(tooltip)

    def _set_base_size(self, add_w=0.0, add_h=0.0):
        """
        Sets the initial base size for the node.

        Args:
            add_w (float): add additional width.
            add_h (float): add additional height.
        """
        self._width, self._height = self.calc_size(add_w, add_h)
        if self._width < NodeEnum.WIDTH.value:
            self._width = NodeEnum.WIDTH.value
        if self._height < NodeEnum.HEIGHT.value:
            self._height = NodeEnum.HEIGHT.value

    def _set_text_color(self, color):
        """
        set text color.

        Args:
            color (tuple): color value in (r, g, b, a).
        """
        text_color = QtGui.QColor(*color)
        for port, text in self._input_items.items():
            text.setDefaultTextColor(text_color)
        for port, text in self._output_items.items():
            text.setDefaultTextColor(text_color)
        self._text_item.setDefaultTextColor(text_color)

    def activate_pipes(self):
        """
        active pipe color.
        """
        ports = self.inputs + self.outputs
        for port in ports:
            for pipe in port.connected_pipes:
                pipe.activate()

    def highlight_pipes(self):
        """
        Highlight pipe color.
        """
        ports = self.inputs + self.outputs
        for port in ports:
            for pipe in port.connected_pipes:
                pipe.highlight()

    def reset_pipes(self):
        """
        Reset all the pipe colors.
        """
        ports = self.inputs + self.outputs
        for port in ports:
            for pipe in port.connected_pipes:
                pipe.reset()

    def _calc_size_horizontal(self):
        # width, height from node name text.
        text_w = self._text_item.boundingRect().width()
        text_h = self._text_item.boundingRect().height()

        # width, height from node ports.
        port_width = 0.0
        p_input_text_width = 0.0
        p_output_text_width = 0.0
        p_input_height = 0.0
        p_output_height = 0.0
        for port, text in self._input_items.items():
            if not port.isVisible():
                continue
            if not port_width:
                port_width = port.boundingRect().width()
            t_width = text.boundingRect().width()
            if text.isVisible() and t_width > p_input_text_width:
                p_input_text_width = text.boundingRect().width()
            p_input_height += port.boundingRect().height()
        for port, text in self._output_items.items():
            if not port.isVisible():
                continue
            if not port_width:
                port_width = port.boundingRect().width()
            t_width = text.boundingRect().width()
            if text.isVisible() and t_width > p_output_text_width:
                p_output_text_width = text.boundingRect().width()
            p_output_height += port.boundingRect().height()

        port_text_width = p_input_text_width + p_output_text_width

        # width, height from node embedded widgets.
        widget_width = 0.0
        widget_height = 0.0
        for widget in self._widgets.values():
            if not widget.isVisible():
                continue
            w_width = widget.boundingRect().width()
            w_height = widget.boundingRect().height()
            if w_width > widget_width:
                widget_width = w_width
            widget_height += w_height

        side_padding = 0.0
        if all([widget_width, p_input_text_width, p_output_text_width]):
            port_text_width = max([p_input_text_width, p_output_text_width])
            port_text_width *= 2
        elif widget_width:
            side_padding = 10

        width = port_width + max([text_w, port_text_width]) + side_padding
        height = max([text_h, p_input_height, p_output_height, widget_height])
        if widget_width:
            # add additional width for node widget.
            width += widget_width
        if widget_height:
            # add bottom margin for node widget.
            height += 4.0
        height *= 1.05
        return width, height

    def _calc_size_vertical(self):
        p_input_width = 0.0
        p_output_width = 0.0
        p_input_height = 0.0
        p_output_height = 0.0
        for port in self._input_items.keys():
            if port.isVisible():
                p_input_width += port.boundingRect().width()
                if not p_input_height:
                    p_input_height = port.boundingRect().height()
        for port in self._output_items.keys():
            if port.isVisible():
                p_output_width += port.boundingRect().width()
                if not p_output_height:
                    p_output_height = port.boundingRect().height()

        widget_width = 0.0
        widget_height = 0.0
        for widget in self._widgets.values():
            if not widget.isVisible():
                continue
            if widget.boundingRect().width() > widget_width:
                widget_width = widget.boundingRect().width()
            widget_height += widget.boundingRect().height()

        width = max([p_input_width, p_output_width, widget_width])
        height = p_input_height + p_output_height + widget_height
        return width, height

    def calc_size(self, add_w=0.0, add_h=0.0):
        """
        Calculates the minimum node size.

        Args:
            add_w (float): additional width.
            add_h (float): additional height.

        Returns:
            tuple(float, float): width, height.
        """
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            width, height = self._calc_size_horizontal()
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            width, height = self._calc_size_vertical()
        else:
            raise RuntimeError("Node graph layout direction not valid!")

        # additional width, height.
        width += add_w
        height += add_h
        return width, height

    def _align_icon_horizontal(self, h_offset, v_offset):
        icon_rect = self._icon_item.boundingRect()
        text_rect = self._text_item.boundingRect()
        x = self.boundingRect().left() + 2.0
        y = text_rect.center().y() - (icon_rect.height() / 2)
        self._icon_item.setPos(x + h_offset, y + v_offset)

    def _align_icon_vertical(self, h_offset, v_offset):
        center_y = self.boundingRect().center().y()
        icon_rect = self._icon_item.boundingRect()
        text_rect = self._text_item.boundingRect()
        x = self.boundingRect().right() + h_offset
        y = center_y - text_rect.height() - (icon_rect.height() / 2) + v_offset
        self._icon_item.setPos(x, y)

    def align_icon(self, h_offset=0.0, v_offset=0.0):
        """
        Align node icon to the default top left of the node.

        Args:
            v_offset (float): additional vertical offset.
            h_offset (float): additional horizontal offset.
        """
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            self._align_icon_horizontal(h_offset, v_offset)
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            self._align_icon_vertical(h_offset, v_offset)
        else:
            raise RuntimeError("Node graph layout direction not valid!")

    def _align_label_horizontal(self, h_offset, v_offset):
        rect = self.boundingRect()
        text_rect = self._text_item.boundingRect()
        x = rect.center().x() - (text_rect.width() / 2)
        self._text_item.setPos(x + h_offset, rect.y() + v_offset)

    def _align_label_vertical(self, h_offset, v_offset):
        rect = self._text_item.boundingRect()
        x = self.boundingRect().right() + h_offset
        y = self.boundingRect().center().y() - (rect.height() / 2) + v_offset
        self.text_item.setPos(x, y)

    def align_label(self, h_offset=0.0, v_offset=0.0):
        """
        Center node label text to the top of the node.

        Args:
            v_offset (float): vertical offset.
            h_offset (float): horizontal offset.
        """
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            self._align_label_horizontal(h_offset, v_offset)
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            self._align_label_vertical(h_offset, v_offset)
        else:
            raise RuntimeError("Node graph layout direction not valid!")

    def _align_widgets_horizontal(self, v_offset):
        if not self._widgets:
            return
        rect = self.boundingRect()
        y = rect.y() + v_offset
        inputs = [p for p in self.inputs if p.isVisible()]
        outputs = [p for p in self.outputs if p.isVisible()]
        for widget in self._widgets.values():
            if not widget.isVisible():
                continue
            widget_rect = widget.boundingRect()
            if not inputs:
                x = rect.left() + 10
                widget.widget().setTitleAlign("left")
            elif not outputs:
                x = rect.right() - widget_rect.width() - 10
                widget.widget().setTitleAlign("right")
            else:
                x = rect.center().x() - (widget_rect.width() / 2)
                widget.widget().setTitleAlign("center")
            widget.setPos(x, y)
            y += widget_rect.height()

    def _align_widgets_vertical(self, v_offset):
        if not self._widgets:
            return
        rect = self.boundingRect()
        y = rect.center().y() + v_offset
        widget_height = 0.0
        for widget in self._widgets.values():
            if not widget.isVisible():
                continue
            widget_rect = widget.boundingRect()
            widget_height += widget_rect.height()
        y -= widget_height / 2

        for widget in self._widgets.values():
            if not widget.isVisible():
                continue
            widget_rect = widget.boundingRect()
            x = rect.center().x() - (widget_rect.width() / 2)
            widget.widget().setTitleAlign("center")
            widget.setPos(x, y)
            y += widget_rect.height()

    def align_widgets(self, v_offset=0.0):
        """
        Align node widgets to the default center of the node.

        Args:
            v_offset (float): vertical offset.
        """
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            self._align_widgets_horizontal(v_offset)
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            self._align_widgets_vertical(v_offset)
        else:
            raise RuntimeError("Node graph layout direction not valid!")

    def _align_ports_horizontal(self, v_offset):
        width = self._width
        txt_offset = PortEnum.CLICK_FALLOFF.value - 2
        spacing = 1

        # adjust input position
        inputs = [p for p in self.inputs if p.isVisible()]
        if inputs:
            port_width = inputs[0].boundingRect().width()
            port_height = inputs[0].boundingRect().height()
            port_x = (port_width / 2) * -1
            port_y = v_offset
            for port in inputs:
                port.setPos(port_x, port_y)
                port_y += port_height + spacing
        # adjust input text position
        for port, text in self._input_items.items():
            if port.isVisible():
                txt_x = port.boundingRect().width() / 2 - txt_offset
                text.setPos(txt_x, port.y() - 1.5)

        # adjust output position
        outputs = [p for p in self.outputs if p.isVisible()]
        if outputs:
            port_width = outputs[0].boundingRect().width()
            port_height = outputs[0].boundingRect().height()
            port_x = width - (port_width / 2)
            port_y = v_offset
            for port in outputs:
                port.setPos(port_x, port_y)
                port_y += port_height + spacing
        # adjust output text position
        for port, text in self._output_items.items():
            if port.isVisible():
                txt_width = text.boundingRect().width() - txt_offset
                txt_x = port.x() - txt_width
                text.setPos(txt_x, port.y() - 1.5)

    def _align_ports_vertical(self, v_offset):
        # adjust input position
        inputs = [p for p in self.inputs if p.isVisible()]
        if inputs:
            port_width = inputs[0].boundingRect().width()
            port_height = inputs[0].boundingRect().height()
            half_width = port_width / 2
            delta = self._width / (len(inputs) + 1)
            port_x = delta
            port_y = (port_height / 2) * -1
            for port in inputs:
                port.setPos(port_x - half_width, port_y)
                port_x += delta

        # adjust output position
        outputs = [p for p in self.outputs if p.isVisible()]
        if outputs:
            port_width = outputs[0].boundingRect().width()
            port_height = outputs[0].boundingRect().height()
            half_width = port_width / 2
            delta = self._width / (len(outputs) + 1)
            port_x = delta
            port_y = self._height - (port_height / 2)
            for port in outputs:
                port.setPos(port_x - half_width, port_y)
                port_x += delta

    def align_ports(self, v_offset=0.0):
        """
        Align input, output ports in the node layout.

        Args:
            v_offset (float): port vertical offset.
        """
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            self._align_ports_horizontal(v_offset)
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            self._align_ports_vertical(v_offset)
        else:
            raise RuntimeError("Node graph layout direction not valid!")

    def _draw_node_horizontal(self):
        height = self._text_item.boundingRect().height() + 4.0

        # update port text items in visibility.
        for port, text in self._input_items.items():
            if port.isVisible():
                text.setVisible(port.display_name)
        for port, text in self._output_items.items():
            if port.isVisible():
                text.setVisible(port.display_name)

        # setup initial base size.
        self._set_base_size(add_h=height)
        # set text color when node is initialized.
        self._set_text_color(self.text_color)
        # set the tooltip
        self._tooltip_disable(self.disabled)

        # --- set the initial node layout ---
        # (do all the graphic item layout offsets here)

        # align label text
        self.align_label()
        # align icon
        self.align_icon(h_offset=2.0, v_offset=1.0)
        # arrange input and output ports.
        self.align_ports(v_offset=height)
        # arrange node widgets
        self.align_widgets(v_offset=height)

        self.update()

    def _draw_node_vertical(self):
        # hide the port text items in vertical layout.
        for port, text in self._input_items.items():
            text.setVisible(False)
        for port, text in self._output_items.items():
            text.setVisible(False)

        # setup initial base size.
        self._set_base_size()
        # set text color when node is initialized.
        self._set_text_color(self.text_color)
        # set the tooltip
        self._tooltip_disable(self.disabled)

        # --- setup node layout ---
        # (do all the graphic item layout offsets here)

        # align label text
        self.align_label(h_offset=6)
        # align icon
        self.align_icon(h_offset=6, v_offset=4)
        # arrange input and output ports.
        self.align_ports()
        # arrange node widgets
        self.align_widgets()

        self.update()

    def draw_node(self):
        """
        Re-draw the node item in the scene with proper
        calculated size and widgets aligned.
        """
        if self.layout_direction is LayoutDirectionEnum.HORIZONTAL.value:
            self._draw_node_horizontal()
        elif self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            self._draw_node_vertical()
        else:
            raise RuntimeError("Node graph layout direction not valid!")

    def post_init(self, viewer=None, pos=None):
        """
        Called after node has been added into the scene.
        Adjust the node layout and form after the node has been added.

        Args:
            viewer (NodeGraphQt.widgets.viewer.NodeViewer): not used
            pos (tuple): cursor position.
        """
        self.draw_node()

        # set initial node position.
        if pos:
            self.xy_pos = pos

    def auto_switch_mode(self):
        """
        Decide whether to draw the node with proxy mode.
        (this is called at the start in the "self.paint()" function.)
        """
        if ITEM_CACHE_MODE is QtWidgets.QGraphicsItem.ItemCoordinateCache:
            return

        rect = self.sceneBoundingRect()
        l = self.viewer().mapToGlobal(self.viewer().mapFromScene(rect.topLeft()))
        r = self.viewer().mapToGlobal(self.viewer().mapFromScene(rect.topRight()))
        # width is the node width in screen
        width = r.x() - l.x()

        self.set_proxy_mode(width < self._proxy_mode_threshold)

    def set_proxy_mode(self, mode):
        """
        Set whether to draw the node with proxy mode.
        (proxy mode toggles visibility for some qgraphic items in the node.)

        Args:
            mode (bool): true to enable proxy mode.
        """
        if mode is self._proxy_mode:
            return
        self._proxy_mode = mode

        visible = not mode

        # disable overlay item.
        self._x_item.proxy_mode = self._proxy_mode

        # node widget visibility.
        for w in self._widgets.values():
            w.widget().setVisible(visible)

        # port text is not visible in vertical layout.
        if self.layout_direction is LayoutDirectionEnum.VERTICAL.value:
            port_text_visible = False
        else:
            port_text_visible = visible

        # input port text visibility.
        for port, text in self._input_items.items():
            if port.display_name:
                text.setVisible(port_text_visible)

        # output port text visibility.
        for port, text in self._output_items.items():
            if port.display_name:
                text.setVisible(port_text_visible)

        self._text_item.setVisible(visible)
        self._icon_item.setVisible(visible)

    @property
    def icon(self):
        return self._properties["icon"]

    @icon.setter
    def icon(self, path=None):
        self._properties["icon"] = path
        path = path or ICON_NODE_BASE
        pixmap = QtGui.QPixmap(path)
        if pixmap.size().height() > NodeEnum.ICON_SIZE.value:
            pixmap = pixmap.scaledToHeight(
                NodeEnum.ICON_SIZE.value, QtCore.Qt.SmoothTransformation
            )
        self._icon_item.setPixmap(pixmap)
        if self.scene():
            self.post_init()

        self.update()

    @AbstractNodeItem.layout_direction.setter
    def layout_direction(self, value=0):
        AbstractNodeItem.layout_direction.fset(self, value)
        self.draw_node()

    @AbstractNodeItem.width.setter
    def width(self, width=0.0):
        w, h = self.calc_size()
        width = width if width > w else w
        AbstractNodeItem.width.fset(self, width)

    @AbstractNodeItem.height.setter
    def height(self, height=0.0):
        w, h = self.calc_size()
        h = 70 if h < 70 else h
        height = height if height > h else h
        AbstractNodeItem.height.fset(self, height)

    @AbstractNodeItem.disabled.setter
    def disabled(self, state=False):
        AbstractNodeItem.disabled.fset(self, state)
        for n, w in self._widgets.items():
            w.widget().setDisabled(state)
        self._tooltip_disable(state)
        self._x_item.setVisible(state)

    @AbstractNodeItem.selected.setter
    def selected(self, selected=False):
        AbstractNodeItem.selected.fset(self, selected)
        if selected:
            self.highlight_pipes()

    @AbstractNodeItem.name.setter
    def name(self, name=""):
        AbstractNodeItem.name.fset(self, name)
        if name == self._text_item.toPlainText():
            return
        self._text_item.setPlainText(name)
        if self.scene():
            self.align_label()
        self.update()

    @AbstractNodeItem.color.setter
    def color(self, color=(100, 100, 100, 255)):
        AbstractNodeItem.color.fset(self, color)
        if self.scene():
            self.scene().update()
        self.update()

    @AbstractNodeItem.text_color.setter
    def text_color(self, color=(100, 100, 100, 255)):
        AbstractNodeItem.text_color.fset(self, color)
        self._set_text_color(color)
        self.update()

    @property
    def text_item(self):
        """
        Get the node name text qgraphics item.

        Returns:
            NodeTextItem: node text object.
        """
        return self._text_item

    @property
    def icon_item(self):
        """
        Get the node icon qgraphics item.

        Returns:
            QtWidgets.QGraphicsPixmapItem: node icon object.
        """
        return self._icon_item

    @property
    def inputs(self):
        """
        Returns:
            list[PortItem]: input port graphic items.
        """
        return list(self._input_items.keys())

    @property
    def outputs(self):
        """
        Returns:
            list[PortItem]: output port graphic items.
        """
        return list(self._output_items.keys())

    def _add_port(self, port):
        """
        Adds a port qgraphics item into the node.

        Args:
            port (PortItem): port item.

        Returns:
            PortItem: port qgraphics item.
        """
        text = QtWidgets.QGraphicsTextItem(port.name, self)
        text.font().setPointSize(8)
        text.setFont(text.font())
        text.setVisible(port.display_name)
        text.setCacheMode(ITEM_CACHE_MODE)
        if port.port_type == PortTypeEnum.IN.value:
            self._input_items[port] = text
        elif port.port_type == PortTypeEnum.OUT.value:
            self._output_items[port] = text
        if self.scene():
            self.post_init()
        return port

    def add_input(
        self,
        name="input",
        multi_port=False,
        display_name=True,
        locked=False,
        painter_func=None,
    ):
        """
        Adds a port qgraphics item into the node with the "port_type" set as
        IN_PORT.

        Args:
            name (str): name for the port.
            multi_port (bool): allow multiple connections.
            display_name (bool): display the port name.
            locked (bool): locked state.
            painter_func (function): custom paint function.

        Returns:
            PortItem: input port qgraphics item.
        """
        if painter_func:
            port = CustomPortItem(self, painter_func)
        else:
            port = PortItem(self)
        port.name = name
        port.port_type = PortTypeEnum.IN.value
        port.multi_connection = multi_port
        port.display_name = display_name
        port.locked = locked
        return self._add_port(port)

    def add_output(
        self,
        name="output",
        multi_port=False,
        display_name=True,
        locked=False,
        painter_func=None,
    ):
        """
        Adds a port qgraphics item into the node with the "port_type" set as
        OUT_PORT.

        Args:
            name (str): name for the port.
            multi_port (bool): allow multiple connections.
            display_name (bool): display the port name.
            locked (bool): locked state.
            painter_func (function): custom paint function.

        Returns:
            PortItem: output port qgraphics item.
        """
        if painter_func:
            port = CustomPortItem(self, painter_func)
        else:
            port = PortItem(self)
        port.name = name
        port.port_type = PortTypeEnum.OUT.value
        port.multi_connection = multi_port
        port.display_name = display_name
        port.locked = locked
        return self._add_port(port)

    def _delete_port(self, port, text):
        """
        Removes port item and port text from node.

        Args:
            port (PortItem): port object.
            text (QtWidgets.QGraphicsTextItem): port text object.
        """
        port.setParentItem(None)
        text.setParentItem(None)
        self.scene().removeItem(port)
        self.scene().removeItem(text)
        del port
        del text

    def delete_input(self, port):
        """
        Remove input port from node.

        Args:
            port (PortItem): port object.
        """
        self._delete_port(port, self._input_items.pop(port))

    def delete_output(self, port):
        """
        Remove output port from node.

        Args:
            port (PortItem): port object.
        """
        self._delete_port(port, self._output_items.pop(port))

    def get_input_text_item(self, port_item):
        """
        Args:
            port_item (PortItem): port item.

        Returns:
            QGraphicsTextItem: graphic item used for the port text.
        """
        return self._input_items[port_item]

    def get_output_text_item(self, port_item):
        """
        Args:
            port_item (PortItem): port item.

        Returns:
            QGraphicsTextItem: graphic item used for the port text.
        """
        return self._output_items[port_item]

    @property
    def widgets(self):
        return self._widgets.copy()

    def add_widget(self, widget):
        self._widgets[widget.get_name()] = widget

    def get_widget(self, name):
        widget = self._widgets.get(name)
        if widget:
            return widget
        raise NodeWidgetError('node has no widget "{}"'.format(name))

    def has_widget(self, name):
        return name in self._widgets.keys()

    def from_dict(self, node_dict):
        super(NodeItem, self).from_dict(node_dict)
        custom_prop = node_dict.get("custom") or {}
        for prop_name, value in custom_prop.items():
            prop_widget = self._widgets.get(prop_name)
            if prop_widget:
                prop_widget.set_value(value)
